<?php

namespace Drupal\salesforce_eca\Commands;

use Drush\Commands\DrushCommands;
use Drupal\Core\Action\ActionManager;
use Drupal\eca\Token\TokenServices;

/**
 * Drush commands for testing Salesforce ECA.
 */
class SalesforceEcaCommands extends DrushCommands {

  /**
   * The action manager.
   *
   * @var \Drupal\Core\Action\ActionManager
   */
  protected $actionManager;

  /**
   * The token services.
   *
   * @var \Drupal\eca\Token\TokenServices
   */
  protected $tokenServices;

  /**
   * Constructs a new SalesforceEcaCommands object.
   */
  public function __construct(ActionManager $action_manager, TokenServices $token_services) {
    parent::__construct();
    $this->actionManager = $action_manager;
    $this->tokenServices = $token_services;
  }

  /**
   * Test the Call Salesforce API action.
   *
   * @command salesforce-eca:test-api
   * @aliases sfeca-test
   * @param string $endpoint The API endpoint to call
   * @param string $method The HTTP method (GET, POST, etc.)
   * @option body Request body for POST/PUT/PATCH
   * @option headers Custom headers (format: "Header1: Value1|Header2: Value2")
   * @usage salesforce-eca:test-api "sobjects/Account/describe" GET
   *   Get Account object metadata
   * @usage salesforce-eca:test-api "sobjects/Lead" POST --body='{"LastName":"Test","Company":"Test Co"}'
   *   Create a new Lead
   */
  public function testApi($endpoint, $method = 'GET', array $options = ['body' => '', 'headers' => '']) {
    $this->output()->writeln("Testing Salesforce Call API action...");
    $this->output()->writeln("Endpoint: $endpoint");
    $this->output()->writeln("Method: $method");

    try {
      // Parse headers if provided.
      $headers = '';
      if (!empty($options['headers'])) {
        $headers = str_replace('|', "\n", $options['headers']);
      }

      // Create and execute the action.
      $action = $this->actionManager->createInstance('salesforce_call_api', [
        'api_endpoint' => $endpoint,
        'http_method' => $method,
        'request_body' => $options['body'],
        'custom_headers' => $headers,
        'return_full_response' => TRUE,
        'result_token' => 'drush_test_result',
      ]);

      $action->execute();

      // Get the token data.
      $token_data = $this->tokenServices->getTokenData('drush_test_result');

      // Display results.
      $this->output()->writeln("\n=== RESPONSE ===");
      $this->output()->writeln("Success: " . ($token_data['success'] ? 'YES' : 'NO'));
      $this->output()->writeln("Status Code: " . $token_data['status_code']);
      $this->output()->writeln("Reason: " . $token_data['reason_phrase']);
      
      if (!empty($token_data['api_usage'])) {
        $this->output()->writeln("API Usage: " . $token_data['api_usage']);
      }

      if (!empty($token_data['error'])) {
        $this->output()->writeln("Error: " . $token_data['error']);
      }

      // Display response data.
      if (!empty($token_data['data'])) {
        $this->output()->writeln("\n=== RESPONSE DATA ===");
        
        // For query results, show record count.
        if (isset($token_data['data']['totalSize'])) {
          $this->output()->writeln("Total Records: " . $token_data['data']['totalSize']);
          $this->output()->writeln("Query Complete: " . ($token_data['data']['done'] ? 'YES' : 'NO'));
        }
        
        // Show first few items of response.
        $this->output()->writeln("Data Preview:");
        $preview = json_encode($token_data['data'], JSON_PRETTY_PRINT);
        if (strlen($preview) > 1000) {
          $preview = substr($preview, 0, 1000) . "\n... (truncated)";
        }
        $this->output()->writeln($preview);
      }

      // Show available fields for SObject responses.
      if (!empty($token_data['fields'])) {
        $this->output()->writeln("\n=== AVAILABLE FIELDS ===");
        $this->output()->writeln(implode(', ', array_keys($token_data['fields'])));
      }

    }
    catch (\Exception $e) {
      $this->output()->writeln("ERROR: " . $e->getMessage());
    }
  }

  /**
   * Run a comprehensive test suite for the Call API action.
   *
   * @command salesforce-eca:test-suite
   * @aliases sfeca-suite
   * @usage salesforce-eca:test-suite
   *   Run all API tests
   */
  public function testSuite() {
    $this->output()->writeln("Running Salesforce ECA Call API Test Suite...\n");

    $tests = [
      [
        'name' => 'Get API Resources',
        'endpoint' => '',
        'method' => 'GET',
      ],
      [
        'name' => 'List Objects',
        'endpoint' => 'sobjects',
        'method' => 'GET',
      ],
      [
        'name' => 'Describe Account',
        'endpoint' => 'sobjects/Account/describe',
        'method' => 'GET',
      ],
      [
        'name' => 'Query Accounts',
        'endpoint' => 'query?q=' . urlencode('SELECT Id, Name FROM Account LIMIT 5'),
        'method' => 'GET',
      ],
      [
        'name' => 'Test Error Handling',
        'endpoint' => 'sobjects/InvalidObject',
        'method' => 'GET',
      ],
    ];

    foreach ($tests as $test) {
      $this->output()->writeln("=== TEST: {$test['name']} ===");
      
      try {
        $action = $this->actionManager->createInstance('salesforce_call_api', [
          'api_endpoint' => $test['endpoint'],
          'http_method' => $test['method'],
          'result_token' => 'test_result',
        ]);

        $action->execute();
        $token_data = $this->tokenServices->getTokenData('test_result');

        $status = $token_data['success'] ? 'PASS' : 'FAIL';
        $this->output()->writeln("Result: $status (Status Code: {$token_data['status_code']})");
        
        if (!$token_data['success']) {
          $this->output()->writeln("Error: " . ($token_data['error'] ?? 'Unknown error'));
        }
      }
      catch (\Exception $e) {
        $this->output()->writeln("Result: EXCEPTION - " . $e->getMessage());
      }
      
      $this->output()->writeln("");
    }
  }

}
