<?php

namespace Drupal\salesforce_eca\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;

/**
 * Test controller for Salesforce ECA Call API action.
 */
class SalesforceEcaTestController extends ControllerBase implements ContainerInjectionInterface {

  /**
   * The ECA processor.
   *
   * @var \Drupal\eca\Processor
   */
  protected $processor;

  /**
   * The action manager.
   *
   * @var \Drupal\Core\Action\ActionManager
   */
  protected $actionManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->processor = $container->get('eca.processor');
    $instance->actionManager = $container->get('plugin.manager.action');
    return $instance;
  }

  /**
   * Test the Call API action.
   */
  public function testCallApi() {
    $results = [];

    try {
      // Test 1: Get Salesforce Resources
      $action = $this->actionManager->createInstance('salesforce_call_api', [
        'api_endpoint' => '',  // Empty path returns resources
        'http_method' => 'GET',
        'result_token' => 'test_resources',
      ]);
      
      $action->execute();
      $token_data = \Drupal::service('eca.token_services')->getTokenData('test_resources');
      
      $results['resources'] = [
        'success' => $token_data['success'] ?? 0,
        'status_code' => $token_data['status_code'] ?? 'N/A',
        'has_data' => !empty($token_data['data']),
        'api_usage' => $token_data['api_usage'] ?? 'N/A',
      ];

      // Test 2: Get Account Object Describe
      $action = $this->actionManager->createInstance('salesforce_call_api', [
        'api_endpoint' => 'sobjects/Account/describe',
        'http_method' => 'GET',
        'result_token' => 'test_describe',
      ]);
      
      $action->execute();
      $token_data = \Drupal::service('eca.token_services')->getTokenData('test_describe');
      
      $results['describe'] = [
        'success' => $token_data['success'] ?? 0,
        'status_code' => $token_data['status_code'] ?? 'N/A',
        'object_name' => $token_data['data']['name'] ?? 'N/A',
        'field_count' => isset($token_data['data']['fields']) ? count($token_data['data']['fields']) : 0,
      ];

      // Test 3: Simple SOQL Query
      $action = $this->actionManager->createInstance('salesforce_call_api', [
        'api_endpoint' => 'query?q=' . urlencode('SELECT Id, Name FROM Account LIMIT 3'),
        'http_method' => 'GET',
        'result_token' => 'test_query',
      ]);
      
      $action->execute();
      $token_data = \Drupal::service('eca.token_services')->getTokenData('test_query');
      
      $results['query'] = [
        'success' => $token_data['success'] ?? 0,
        'status_code' => $token_data['status_code'] ?? 'N/A',
        'total_size' => $token_data['data']['totalSize'] ?? 0,
        'done' => $token_data['data']['done'] ?? FALSE,
      ];

      // Test 4: Error Handling (Invalid endpoint)
      $action = $this->actionManager->createInstance('salesforce_call_api', [
        'api_endpoint' => 'sobjects/InvalidObjectName',
        'http_method' => 'GET',
        'result_token' => 'test_error',
      ]);
      
      $action->execute();
      $token_data = \Drupal::service('eca.token_services')->getTokenData('test_error');
      
      $results['error_test'] = [
        'success' => $token_data['success'] ?? 0,
        'status_code' => $token_data['status_code'] ?? 'N/A',
        'error' => $token_data['error'] ?? 'N/A',
        'reason_phrase' => $token_data['reason_phrase'] ?? 'N/A',
      ];

    }
    catch (\Exception $e) {
      $results['exception'] = $e->getMessage();
    }

    return new JsonResponse($results);
  }

}
