<?php

namespace Drupal\salesforce_eca\EventSubscriber;

use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\salesforce\Event\SalesforceEvents;
use Drupal\salesforce_mapping\Event\SalesforcePullEnqueueEvent;
use Drupal\salesforce_mapping\Event\SalesforcePullEntityValueEvent;
use Drupal\salesforce_mapping\Event\SalesforcePullEvent;
use Drupal\salesforce_mapping\Event\SalesforcePushAllowedEvent;
use Drupal\salesforce_mapping\Event\SalesforcePushOpEvent;
use Drupal\salesforce_mapping\Event\SalesforcePushParamsEvent;
use Drupal\salesforce_mapping\Event\SalesforceQueryEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Subscribes to Salesforce events for logging and debugging.
 *
 * Note: This subscriber only logs events. ECA events are triggered
 * automatically by the ECA event plugins when Salesforce events occur.
 */
class SalesforceEcaEventSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * Pull enqueue event callback.
   *
   * @param \Drupal\salesforce_mapping\Event\SalesforcePullEnqueueEvent $event
   *   The event.
   */
  public function pullEnqueue(SalesforcePullEnqueueEvent $event) {
    $this->loggerFactory->get('salesforce_eca')->debug('Salesforce pull enqueue event fired for mapping: @mapping', [
      '@mapping' => $event->getMapping()->id(),
    ]);
  }

  /**
   * Pull entity value event callback.
   *
   * @param \Drupal\salesforce_mapping\Event\SalesforcePullEntityValueEvent $event
   *   The event.
   */
  public function pullEntityValue(SalesforcePullEntityValueEvent $event) {
    $entity = $event->getEntity();
    $this->loggerFactory->get('salesforce_eca')->debug('Salesforce pull entity value event fired for @entity_type:@entity_id with mapping @mapping', [
      '@entity_type' => $entity ? $entity->getEntityTypeId() : 'unknown',
      '@entity_id' => $entity && $entity->id() ? (string) $entity->id() : 'new',
      '@mapping' => $event->getMapping()->id(),
    ]);
  }

  /**
   * SalesforcePushMappingObject event callback.
   *
   * @param \Drupal\salesforce_mapping\Event\SalesforcePushOpEvent $event
   *   The event.
   */
  public function pushMappingObject(SalesforcePushOpEvent $event) {
    $entity = $event->getEntity();
    $this->loggerFactory->get('salesforce_eca')->debug('Salesforce push mapping object event fired for @entity_type:@entity_id with mapping @mapping', [
      '@entity_type' => $entity ? $entity->getEntityTypeId() : 'unknown',
      '@entity_id' => $entity && $entity->id() ? (string) $entity->id() : 'new',
      '@mapping' => $event->getMapping()->id(),
    ]);
  }

  /**
   * The logger factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * Constructs a new SalesforceEcaEventSubscriber.
   *
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory.
   */
  public function __construct(LoggerChannelFactoryInterface $logger_factory) {
    $this->loggerFactory = $logger_factory;
  }

  /**
   * SalesforcePushAllowedEvent callback.
   *
   * @param \Drupal\salesforce_mapping\Event\SalesforcePushAllowedEvent $event
   *   The push allowed event.
   */
  public function pushAllowed(SalesforcePushAllowedEvent $event) {
    $entity = $event->getEntity();
    $this->loggerFactory->get('salesforce_eca')->debug('Salesforce push allowed event fired for @entity_type:@entity_id with mapping @mapping', [
      '@entity_type' => $entity ? $entity->getEntityTypeId() : 'unknown',
      '@entity_id' => $entity && $entity->id() ? (string) $entity->id() : 'new',
      '@mapping' => $event->getMapping()->id(),
    ]);
  }

  /**
   * SalesforcePushParamsEvent callback.
   *
   * @param \Drupal\salesforce_mapping\Event\SalesforcePushParamsEvent $event
   *   The event.
   */
  public function pushParamsAlter(SalesforcePushParamsEvent $event) {
    $entity = $event->getEntity();
    $this->loggerFactory->get('salesforce_eca')->debug('Salesforce push params event fired for @entity_type:@entity_id with mapping @mapping', [
      '@entity_type' => $entity ? $entity->getEntityTypeId() : 'unknown',
      '@entity_id' => $entity && $entity->id() ? (string) $entity->id() : 'new',
      '@mapping' => $event->getMapping()->id(),
    ]);
  }

  /**
   * SalesforcePushSuccess event callback.
   *
   * @param \Drupal\salesforce_mapping\Event\SalesforcePushParamsEvent $event
   *   The event.
   */
  public function pushSuccess(SalesforcePushParamsEvent $event) {
    $entity = $event->getEntity();
    $this->loggerFactory->get('salesforce_eca')->debug('Salesforce push success event fired for @entity_type:@entity_id with mapping @mapping', [
      '@entity_type' => $entity ? $entity->getEntityTypeId() : 'unknown',
      '@entity_id' => $entity && $entity->id() ? (string) $entity->id() : 'unknown',
      '@mapping' => $event->getMapping()->id(),
    ]);
  }

  /**
   * SalesforcePushFail event callback.
   *
   * @param \Drupal\salesforce_mapping\Event\SalesforcePushOpEvent $event
   *   The event.
   */
  public function pushFail(SalesforcePushOpEvent $event) {
    $entity = $event->getEntity();
    $this->loggerFactory->get('salesforce_eca')->debug('Salesforce push fail event fired for @entity_type:@entity_id with mapping @mapping', [
      '@entity_type' => $entity ? $entity->getEntityTypeId() : 'unknown',
      '@entity_id' => $entity && $entity->id() ? (string) $entity->id() : 'unknown',
      '@mapping' => $event->getMapping()->id(),
    ]);
  }

  /**
   * SalesforceQueryEvent pull query alter event callback.
   *
   * @param \Drupal\salesforce_mapping\Event\SalesforceQueryEvent $event
   *   The event.
   */
  public function pullQueryAlter(SalesforceQueryEvent $event) {
    $mapping = $event->getMapping();
    $this->loggerFactory->get('salesforce_eca')
      ->debug('Salesforce pull query event for mapping @mapping object @object_type', [
        '@mapping' => $mapping->id(),
        '@object_type' => $mapping->getSalesforceObjectType(),
      ]);
  }

  /**
   * Pull presave event callback.
   *
   * @param \Drupal\salesforce_mapping\Event\SalesforcePullEvent $event
   *   The event.
   */
  public function pullPresave(SalesforcePullEvent $event) {
    $entity = $event->getEntity();
    $this->loggerFactory->get('salesforce_eca')->debug('Salesforce pull presave event fired for @entity_type:@entity_id with mapping @mapping', [
      '@entity_type' => $entity ? $entity->getEntityTypeId() : 'unknown',
      '@entity_id' => $entity && $entity->id() ? (string) $entity->id() : 'new',
      '@mapping' => $event->getMapping()->id(),
    ]);
  }

  /**
   * Pull prepull event callback.
   *
   * @param \Drupal\salesforce_mapping\Event\SalesforcePullEvent $event
   *   The event.
   */
  public function pullPrepull(SalesforcePullEvent $event) {
    $entity = $event->getEntity();
    $this->loggerFactory->get('salesforce_eca')->debug('Salesforce pull prepull event fired for @entity_type:@entity_id with mapping @mapping', [
      '@entity_type' => $entity ? $entity->getEntityTypeId() : 'unknown',
      '@entity_id' => $entity && $entity->id() ? (string) $entity->id() : 'new',
      '@mapping' => $event->getMapping()->id(),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return [
      SalesforceEvents::PUSH_ALLOWED => 'pushAllowed',
      SalesforceEvents::PUSH_MAPPING_OBJECT => 'pushMappingObject',
      SalesforceEvents::PUSH_PARAMS => 'pushParamsAlter',
      SalesforceEvents::PUSH_SUCCESS => 'pushSuccess',
      SalesforceEvents::PUSH_FAIL => 'pushFail',
      SalesforceEvents::PULL_QUERY => 'pullQueryAlter',
      SalesforceEvents::PULL_ENQUEUE => 'pullEnqueue',
      SalesforceEvents::PULL_PREPULL => 'pullPrepull',
      SalesforceEvents::PULL_ENTITY_VALUE => 'pullEntityValue',
      SalesforceEvents::PULL_PRESAVE => 'pullPresave',
    ];
  }

}
