<?php

namespace Drupal\salesforce_eca\Plugin\Action;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\eca\Plugin\Action\ConfigurableActionBase;
use Drupal\salesforce\Rest\RestClientInterface;
use Drupal\salesforce\Rest\RestResponse;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Call any Salesforce REST API endpoint.
 *
 * @Action(
 *   id = "salesforce_call_api",
 *   label = @Translation("Call Salesforce API"),
 *   description = @Translation("Make a direct call to any Salesforce REST API endpoint with full response access."),
 *   eca_version_introduced = "1.0.0"
 * )
 */
class SalesforceCallApi extends ConfigurableActionBase {

  /**
   * The Salesforce REST client.
   *
   * @var \Drupal\salesforce\Rest\RestClientInterface
   */
  protected RestClientInterface $salesforceClient;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('eca.token_services'),
      $container->get('current_user'),
      $container->get('datetime.time'),
      $container->get('eca.state'),
      $container->get('logger.channel.eca')
    );

    // Inject the Salesforce client service.
    $instance->setSalesforceClient($container->get('salesforce.client'));

    return $instance;
  }

  /**
   * Sets the Salesforce client.
   *
   * @param \Drupal\salesforce\Rest\RestClientInterface $salesforce_client
   *   The Salesforce REST client.
   */
  public function setSalesforceClient(RestClientInterface $salesforce_client): void {
    $this->salesforceClient = $salesforce_client;
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, ?AccountInterface $account = NULL, $return_as_object = FALSE): bool|AccessResultInterface {
    // For ECA actions, we generally allow access since the action execution
    // is controlled by the ECA workflow configuration.
    $access_result = AccessResult::allowed();
    return $return_as_object ? $access_result : $access_result->isAllowed();
  }

  /**
   * {@inheritdoc}
   */
  public function execute($entity = NULL): void {
    $result_token = $this->configuration['result_token'];

    try {
      // Process tokens in configuration values.
      $api_endpoint = $this->tokenService->replaceClear($this->configuration['api_endpoint']);
      $http_method = $this->configuration['http_method'];
      $request_body = $this->tokenService->replaceClear($this->configuration['request_body']);
      $custom_headers = $this->tokenService->replaceClear($this->configuration['custom_headers']);
      $return_full_response = (bool) $this->configuration['return_full_response'];

      // Validate and prepare request body.
      $params = [];
      if (!empty($request_body) && in_array($http_method, ['POST', 'PUT', 'PATCH'])) {
        // Try to decode as JSON first.
        $decoded = json_decode($request_body, TRUE);
        if (json_last_error() === JSON_ERROR_NONE) {
          $params = $decoded;
        }
        else {
          // If not JSON, pass as raw string.
          $params = $request_body;
        }
      }

      // Parse custom headers.
      $headers = [];
      if (!empty($custom_headers)) {
        $header_lines = explode("\n", $custom_headers);
        foreach ($header_lines as $line) {
          $line = trim($line);
          if (!empty($line) && strpos($line, ':') !== FALSE) {
            list($key, $value) = explode(':', $line, 2);
            $headers[trim($key)] = trim($value);
          }
        }
      }

      // Log the API call attempt.
      $this->logger->info('Calling Salesforce API: @method @endpoint', [
        '@method' => $http_method,
        '@endpoint' => $api_endpoint,
      ]);

      // Make the API call with returnObject=TRUE to get full response.
      /** @var \Drupal\salesforce\Rest\RestResponse $response */
      $response = $this->salesforceClient->apiCall(
        $api_endpoint,
        $params,
        $http_method,
        TRUE,
        $headers
      );

      // Extract all response information.
      $status_code = $response->getStatusCode();
      $reason_phrase = $response->getReasonPhrase();
      $all_headers = $response->getHeaders();
      $content_type = $response->getHeaderLine('Content-Type');
      
      // Get API usage from specific header.
      $api_usage_header = $response->getHeaderLine('Sforce-Limit-Info');
      
      // Get the response data.
      $response_data = $response->data;
      
      // Get raw body for non-JSON responses.
      $raw_body = '';
      try {
        $body = $response->getBody();
        if ($body) {
          $body->rewind();
          $raw_body = $body->getContents();
        }
      }
      catch (\Exception $e) {
        // Body might not be available or readable.
        $this->logger->notice('Could not read raw response body: @error', [
          '@error' => $e->getMessage(),
        ]);
      }

      // Determine if response is JSON.
      $is_json = strpos($content_type, 'application/json') !== FALSE;

      // Extract SObject information if applicable.
      $sobject_type = '';
      $record_id = '';
      $field_data = [];
      
      if (is_array($response_data)) {
        // Check for single SObject response pattern.
        if (isset($response_data['attributes']['type'])) {
          $sobject_type = $response_data['attributes']['type'];
        }
        if (isset($response_data['Id'])) {
          $record_id = $response_data['Id'];
        }
        
        // Extract fields (everything except attributes).
        foreach ($response_data as $key => $value) {
          if ($key !== 'attributes') {
            $field_data[$key] = $value;
          }
        }
      }

      // Build comprehensive token data structure.
      $token_data = [
        // Response content (primary data).
        'data' => $response_data,
        'raw_body' => $raw_body,
        
        // HTTP metadata (MDVIP compatibility).
        'status_code' => $status_code,
        'reason_phrase' => $reason_phrase,
        'protocol_version' => $response->getProtocolVersion(),
        
        // Headers (comprehensive access).
        'headers' => $all_headers,
        'content_type' => $content_type,
        'content_length' => $response->getHeaderLine('Content-Length'),
        
        // Salesforce-specific metadata.
        'api_usage' => $api_usage_header,
        
        // Success indicators.
        'success' => ($status_code >= 200 && $status_code < 300) ? 1 : 0,
        'is_json' => $is_json ? 1 : 0,
        
        // SObject helpers (for objectRead-style responses).
        'sobject_type' => $sobject_type,
        'record_id' => $record_id,
        'fields' => $field_data,
        
        // Request information for debugging.
        'request_endpoint' => $api_endpoint,
        'request_method' => $http_method,
      ];

      // Store the token data.
      $this->tokenService->addTokenData($result_token, $token_data);

      // Log success.
      $this->logger->info('Successfully called Salesforce API. Status: @status, Endpoint: @endpoint', [
        '@status' => $status_code,
        '@endpoint' => $api_endpoint,
      ]);

    }
    catch (\Exception $e) {
      // Log the error.
      $this->logger->error('Error calling Salesforce API: @error. Endpoint: @endpoint', [
        '@error' => $e->getMessage(),
        '@endpoint' => $api_endpoint ?? 'unknown',
      ]);

      // Still try to extract response information if available.
      $error_data = [
        'success' => 0,
        'error' => $e->getMessage(),
        'status_code' => 0,
        'reason_phrase' => '',
        'data' => NULL,
        'raw_body' => '',
        'headers' => [],
        'request_endpoint' => $api_endpoint ?? '',
        'request_method' => $http_method ?? 'GET',
      ];

      // If the exception contains a response, extract its data.
      if (method_exists($e, 'getResponse') && $e->getResponse()) {
        $error_response = $e->getResponse();
        $error_data['status_code'] = $error_response->getStatusCode();
        $error_data['reason_phrase'] = $error_response->getReasonPhrase();
        $error_data['headers'] = $error_response->getHeaders();
        
        try {
          $error_body = $error_response->getBody();
          if ($error_body) {
            $error_body->rewind();
            $error_data['raw_body'] = $error_body->getContents();
            
            // Try to decode error response as JSON.
            $decoded_error = json_decode($error_data['raw_body'], TRUE);
            if (json_last_error() === JSON_ERROR_NONE) {
              $error_data['data'] = $decoded_error;
            }
          }
        }
        catch (\Exception $body_exception) {
          // Could not read error response body.
        }
      }

      $this->tokenService->addTokenData($result_token, $error_data);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'api_endpoint' => '',
      'http_method' => 'GET',
      'request_body' => '',
      'custom_headers' => '',
      'return_full_response' => TRUE,
      'result_token' => 'salesforce_api_result',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['api_endpoint'] = [
      '#type' => 'textarea',
      '#title' => $this->t('API Endpoint'),
      '#description' => $this->t('The Salesforce API endpoint to call. You can use tokens in this field.<br>
        <strong>Relative paths</strong> (standard REST API):<br>
        • <code>sobjects/Account/001D000000AbcDE</code> - Get specific record<br>
        • <code>sobjects/Lead/describe</code> - Get object metadata<br>
        • <code>query?q=SELECT Id, Name FROM Account LIMIT 10</code> - SOQL query<br>
        • <code>composite/tree/Account</code> - Composite API<br>
        <strong>Absolute paths</strong> (custom Apex REST endpoints):<br>
        • <code>/services/apexrest/MyRestService</code> - Custom Apex endpoint<br>
        • <code>/services/apexrest/AccountService/validateEmail</code> - Custom method<br>
        <strong>Note:</strong> Paths starting with "/" are treated as absolute URLs.'),
      '#default_value' => $this->configuration['api_endpoint'],
      '#required' => TRUE,
      '#rows' => 3,
    ];

    $form['http_method'] = [
      '#type' => 'select',
      '#title' => $this->t('HTTP Method'),
      '#description' => $this->t('The HTTP method to use for the API call.'),
      '#options' => [
        'GET' => 'GET',
        'POST' => 'POST',
        'PUT' => 'PUT',
        'PATCH' => 'PATCH',
        'DELETE' => 'DELETE',
      ],
      '#default_value' => $this->configuration['http_method'],
      '#required' => TRUE,
    ];

    $form['request_body'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Request Body'),
      '#description' => $this->t('The request body for POST, PUT, or PATCH requests. Should be valid JSON. You can use tokens in this field.<br>
        <strong>Example:</strong><br>
        <code>{<br>
        &nbsp;&nbsp;"Name": "New Account",<br>
        &nbsp;&nbsp;"Type": "Customer",<br>
        &nbsp;&nbsp;"BillingCity": "[entity:field_city]"<br>
        }</code>'),
      '#default_value' => $this->configuration['request_body'],
      '#required' => FALSE,
      '#rows' => 6,
      '#states' => [
        'visible' => [
          ':input[name="http_method"]' => [
            ['value' => 'POST'],
            ['value' => 'PUT'],
            ['value' => 'PATCH'],
          ],
        ],
        'required' => [
          ':input[name="http_method"]' => [
            ['value' => 'POST'],
            ['value' => 'PUT'],
            ['value' => 'PATCH'],
          ],
        ],
      ],
    ];

    $form['custom_headers'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Custom Headers'),
      '#description' => $this->t('Optional custom headers to send with the request. Enter one header per line in the format "Header-Name: Header Value". You can use tokens.<br>
        <strong>Example:</strong><br>
        <code>X-Custom-Header: MyValue<br>
        Accept: application/xml<br>
        X-Request-ID: [system:request_time]</code>'),
      '#default_value' => $this->configuration['custom_headers'],
      '#required' => FALSE,
      '#rows' => 4,
    ];

    $form['return_full_response'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Return full response object'),
      '#description' => $this->t('If checked, the full response object with headers and metadata will be available in tokens. This is recommended for error handling and debugging.'),
      '#default_value' => $this->configuration['return_full_response'],
    ];

    $form['result_token'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Result token name'),
      '#description' => $this->t('The token name to store the API call results. You can access the response data using:<br>
        <strong>Response data:</strong><br>
        • <code>[TOKEN:data]</code> - Parsed response data (JSON decoded)<br>
        • <code>[TOKEN:raw_body]</code> - Raw response body string<br>
        <strong>HTTP metadata (MDVIP compatible):</strong><br>
        • <code>[TOKEN:status_code]</code> - HTTP status code (200, 404, etc.)<br>
        • <code>[TOKEN:reason_phrase]</code> - HTTP reason phrase ("OK", "Not Found")<br>
        • <code>[TOKEN:protocol_version]</code> - HTTP protocol version<br>
        <strong>Headers:</strong><br>
        • <code>[TOKEN:headers]</code> - All response headers<br>
        • <code>[TOKEN:content_type]</code> - Content-Type header<br>
        • <code>[TOKEN:api_usage]</code> - Salesforce API usage (Sforce-Limit-Info)<br>
        <strong>Status flags:</strong><br>
        • <code>[TOKEN:success]</code> - 1 if status code is 2XX, 0 otherwise<br>
        • <code>[TOKEN:is_json]</code> - 1 if response is JSON<br>
        <strong>SObject helpers (for single record responses):</strong><br>
        • <code>[TOKEN:sobject_type]</code> - Object type (e.g., "Account")<br>
        • <code>[TOKEN:record_id]</code> - Record ID if present<br>
        • <code>[TOKEN:fields:FieldName]</code> - Direct field access<br>
        <strong>Error handling:</strong><br>
        • <code>[TOKEN:error]</code> - Error message if request failed<br>
        • <code>[TOKEN:request_endpoint]</code> - The endpoint that was called<br>
        • <code>[TOKEN:request_method]</code> - The HTTP method used'),
      '#default_value' => $this->configuration['result_token'],
      '#required' => TRUE,
      '#eca_token_reference' => TRUE,
      '#element_validate' => [[$this, 'validateTokenName']],
    ];

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * Validates the token name field.
   *
   * @param array $element
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function validateTokenName(array $element, FormStateInterface $form_state): void {
    $value = $element['#value'];
    if (!preg_match('/^[a-zA-Z_][a-zA-Z0-9_]*$/', $value)) {
      $form_state->setError($element, $this->t('Token name must start with a letter or underscore and contain only letters, numbers, and underscores.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->configuration['api_endpoint'] = $form_state->getValue('api_endpoint');
    $this->configuration['http_method'] = $form_state->getValue('http_method');
    $this->configuration['request_body'] = $form_state->getValue('request_body');
    $this->configuration['custom_headers'] = $form_state->getValue('custom_headers');
    $this->configuration['return_full_response'] = $form_state->getValue('return_full_response');
    $this->configuration['result_token'] = $form_state->getValue('result_token');
    parent::submitConfigurationForm($form, $form_state);
  }

}
