<?php

namespace Drupal\salesforce_eca\Plugin\Action;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\eca\Plugin\Action\ConfigurableActionBase;
use Drupal\salesforce\Rest\RestClientInterface;
use Drupal\salesforce\SelectQueryRaw;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Execute a SOQL query against Salesforce.
 *
 * @Action(
 *   id = "salesforce_execute_soql",
 *   label = @Translation("Execute SOQL query"),
 *   description = @Translation("Execute a SOQL query against Salesforce and store the results in a token."),
 *   eca_version_introduced = "1.0.0"
 * )
 */
class SalesforceExecuteSoql extends ConfigurableActionBase {

  /**
   * The Salesforce REST client.
   *
   * @var \Drupal\salesforce\Rest\RestClientInterface
   */
  protected RestClientInterface $salesforceClient;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('eca.token_services'),
      $container->get('current_user'),
      $container->get('datetime.time'),
      $container->get('eca.state'),
      $container->get('logger.channel.eca')
    );

    // Inject the additional Salesforce client service.
    $instance->setSalesforceClient($container->get('salesforce.client'));

    return $instance;
  }

  /**
   * Sets the Salesforce client.
   *
   * @param \Drupal\salesforce\Rest\RestClientInterface $salesforce_client
   *   The Salesforce REST client.
   */
  public function setSalesforceClient(RestClientInterface $salesforce_client): void {
    $this->salesforceClient = $salesforce_client;
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, ?AccountInterface $account = NULL, $return_as_object = FALSE): bool|AccessResultInterface {
    // For ECA actions, we generally allow access since the action execution
    // is controlled by the ECA workflow configuration. The Salesforce
    // connection
    // itself handles authentication, not individual action permissions.
    $access_result = AccessResult::allowed();
    return $return_as_object ? $access_result : $access_result->isAllowed();
  }

  /**
   * {@inheritdoc}
   */
  public function execute($entity = NULL): void {
    // Process tokens in the SOQL query.
    $soql_query = $this->tokenService->replaceClear($this->configuration['soql_query']);

    try {
      // Use SelectQueryRaw for executing raw SOQL queries.
      $query = new SelectQueryRaw($soql_query);
      $result = $this->salesforceClient->query($query);

      // Store the result in a token for use in subsequent actions.
      $token_name = $this->configuration['result_token'];

      // Convert complex objects to serializable data for ECA tokens.
      $records = $result->records();

      $serializable_records = [];

      // Convert each record to an array to make it token-accessible.
      $numeric_index = 0;
      foreach ($records as $record) {

        if (is_object($record)) {
          // Handle Salesforce record objects properly.
          $record_data = [];

          // Try to get record data using common Salesforce patterns.
          if (method_exists($record, 'fields')) {
            $record_data = $record->fields();

          }
          elseif (method_exists($record, 'getData')) {
            $record_data = $record->getData();

          }
          elseif (method_exists($record, 'toArray')) {
            $record_data = $record->toArray();

          }
          elseif (isset($record->fields)) {
            $record_data = (array) $record->fields;

          }
          else {
            // Last resort: try to access object properties.
            $record_data = [];
            $reflection = new \ReflectionObject($record);
            foreach ($reflection->getProperties(\ReflectionProperty::IS_PUBLIC) as $property) {
              $record_data[$property->getName()] = $property->getValue($record);
            }

            // Also try common Salesforce field patterns.
            foreach (['Id', 'Name', 'attributes'] as $field) {
              if (isset($record->$field)) {
                $record_data[$field] = $record->$field;
              }
            }

          }

          $serializable_records[$numeric_index] = $record_data;
        }
        else {
          $serializable_records[$numeric_index] = $record;

        }
        $numeric_index++;
      }

      // Store serializable data structure.
      $token_data = [
        'records' => $serializable_records,
        'size' => $result->size(),
      // Convert boolean to int for better token handling.
        'done' => $result->done() ? 1 : 0,
      // Ensure string value.
        'next_records_url' => $result->nextRecordsUrl() ?: '',
      // Available field names.
        'field_names' => !empty($serializable_records) ? array_keys($serializable_records[0]) : [],
      ];

      $this->tokenService->addTokenData($token_name, $token_data);

      // Log success message.
      $this->logger->info('Successfully executed SOQL query and stored @count results in token "@token".', [
        '@count' => $result->size(),
        '@token' => $token_name,
      ]);
    }
    catch (\Exception $e) {
      // Log the error with more detail.
      $this->logger->error('Error executing SOQL query: @error. Query was: @query', [
        '@error' => $e->getMessage(),
        '@query' => $soql_query,
      ]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'soql_query' => '',
      'result_token' => 'salesforce_soql_result',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['soql_query'] = [
      '#type' => 'textarea',
      '#title' => $this->t('SOQL Query'),
      '#description' => $this->t('Enter the SOQL query to execute. You can use ECA tokens in this field.<br>
        <strong>Example:</strong> SELECT Id, Name, Email FROM Contact WHERE LastModifiedDate > 2024-01-01T00:00:00Z<br>
        <strong>Note:</strong> Large result sets are loaded into memory. Consider using LIMIT clauses for queries that might return many records.'),
      '#default_value' => $this->configuration['soql_query'],
      '#required' => TRUE,
      '#rows' => 5,
    ];

    $form['result_token'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Result token name'),
      '#description' => $this->t('The token name to store the SOQL query results. Only letters, numbers, and underscores are allowed. If you name your token "my_results", you can access the data using:<br>
        • <code>[my_results:size]</code> - Total number of results<br>
        • <code>[my_results:done]</code> - Whether all records have been retrieved (1 or 0)<br>
        • <code>[my_results:field_names]</code> - Array of available field names from the query<br>
        • <code>[my_results:records:0:Name]</code> - First record\'s Name field<br>
        • <code>[my_results:records:0:Id]</code> - First record\'s Id field<br>
        • <code>[my_results:records:1:Name]</code> - Second record\'s Name field<br>
        • <code>[my_results:next_records_url]</code> - URL for next batch (if query has more results)<br>
        <strong>Note:</strong> Use exact Salesforce field names (capitalized like "Name", "Id"). Use <code>field_names</code> token to discover available fields.'),
      '#default_value' => $this->configuration['result_token'],
      '#required' => TRUE,
      '#eca_token_reference' => TRUE,
      '#element_validate' => [[$this, 'validateTokenName']],
    ];

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * Validates the token name field.
   *
   * @param array $element
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function validateTokenName(array $element, FormStateInterface $form_state): void {
    $value = $element['#value'];
    if (!preg_match('/^[a-zA-Z_][a-zA-Z0-9_]*$/', $value)) {
      $form_state->setError($element, $this->t('Token name must start with a letter or underscore and contain only letters, numbers, and underscores.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->configuration['soql_query'] = $form_state->getValue('soql_query');
    $this->configuration['result_token'] = $form_state->getValue('result_token');
    parent::submitConfigurationForm($form, $form_state);
  }

}
