<?php

namespace Drupal\salesforce_eca\Plugin\Action;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\eca\Plugin\Action\ConfigurableActionBase;
use Drupal\salesforce\Rest\RestClientInterface;
use Drupal\salesforce\SObject;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Get field value(s) from a Salesforce object.
 *
 * @Action(
 *   id = "salesforce_get_field_value",
 *   label = @Translation("Get Salesforce field value(s)"),
 *   description = @Translation("Retrieve one or more field values from a Salesforce object. Simpler alternative to SOQL for single record field access."),
 *   eca_version_introduced = "1.0.0"
 * )
 */
class SalesforceGetFieldValue extends ConfigurableActionBase {

  /**
   * The Salesforce REST client.
   *
   * @var \Drupal\salesforce\Rest\RestClientInterface
   */
  protected RestClientInterface $salesforceClient;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('eca.token_services'),
      $container->get('current_user'),
      $container->get('datetime.time'),
      $container->get('eca.state'),
      $container->get('logger.channel.eca')
    );

    // Inject the Salesforce client service.
    $instance->setSalesforceClient($container->get('salesforce.client'));

    return $instance;
  }

  /**
   * Sets the Salesforce client.
   *
   * @param \Drupal\salesforce\Rest\RestClientInterface $salesforce_client
   *   The Salesforce REST client.
   */
  public function setSalesforceClient(RestClientInterface $salesforce_client): void {
    $this->salesforceClient = $salesforce_client;
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, ?AccountInterface $account = NULL, $return_as_object = FALSE): bool|AccessResultInterface {
    // ECA actions generally allow access - workflow controls execution.
    $access_result = AccessResult::allowed();
    return $return_as_object ? $access_result : $access_result->isAllowed();
  }

  /**
   * {@inheritdoc}
   */
  public function execute($entity = NULL): void {
    $result_token = $this->configuration['result_token'];

    try {
      // Process tokens in configuration values.
      $object_type = $this->tokenService->replaceClear($this->configuration['object_type']);
      $record_id = $this->tokenService->replaceClear($this->configuration['record_id']);
      $external_id_field = $this->tokenService->replaceClear($this->configuration['external_id_field']);
      $external_id_value = $this->tokenService->replaceClear($this->configuration['external_id_value']);
      $field_names = $this->tokenService->replaceClear($this->configuration['field_names']);
      $use_external_id = (bool) $this->configuration['use_external_id'];

      // Validate required fields.
      if (empty($object_type)) {
        throw new \Exception('Salesforce object type is required.');
      }

      // Determine which ID method to use.
      if ($use_external_id) {
        if (empty($external_id_field) || empty($external_id_value)) {
          throw new \Exception('External ID field and value are required when using external ID.');
        }
        $this->logger->info('Getting Salesforce fields for @type using external ID @field = @value', [
          '@type' => $object_type,
          '@field' => $external_id_field,
          '@value' => $external_id_value,
        ]);
      }
      else {
        if (empty($record_id)) {
          throw new \Exception('Salesforce record ID is required.');
        }
        $this->logger->info('Getting Salesforce fields for @type/@id', [
          '@type' => $object_type,
          '@id' => $record_id,
        ]);
      }

      // Parse field names into array.
      $requested_fields = [];
      $get_all_fields = FALSE;

      if (empty($field_names) || $field_names === '*') {
        // Get all fields.
        $get_all_fields = TRUE;
      }
      else {
        // Parse comma-separated field names.
        $field_names = str_replace(' ', '', $field_names);
        $requested_fields = explode(',', $field_names);
        $requested_fields = array_filter($requested_fields);
      }

      // Retrieve the Salesforce object.
      try {
        if ($use_external_id) {
          $sf_object = $this->salesforceClient->objectReadbyExternalId(
            $object_type,
            $external_id_field,
            $external_id_value
          );
        }
        else {
          $sf_object = $this->salesforceClient->objectRead(
            $object_type,
            $record_id
          );
        }
      }
      catch (\Exception $e) {
        // Handle specific Salesforce errors.
        $error_message = $e->getMessage();
        if (strpos($error_message, 'NOT_FOUND') !== FALSE) {
          $error_message = 'Record not found in Salesforce.';
        }
        elseif (strpos($error_message, 'INVALID_FIELD') !== FALSE) {
          $error_message = 'Invalid field specified.';
        }
        elseif (strpos($error_message, 'INSUFFICIENT_ACCESS') !== FALSE) {
          $error_message = 'Insufficient access to retrieve this record.';
        }

        throw new \Exception($error_message);
      }

      // Validate we got an SObject.
      if (!$sf_object instanceof SObject) {
        throw new \Exception('Failed to retrieve Salesforce object.');
      }

      // Extract field values.
      $field_values = [];
      $fields_not_found = [];
      $all_fields = $sf_object->fields();

      if ($get_all_fields) {
        // Return all fields.
        $field_values = $all_fields;
      }
      else {
        // Return only requested fields.
        foreach ($requested_fields as $field_name) {
          if ($sf_object->hasField($field_name)) {
            $field_values[$field_name] = $sf_object->field($field_name);
          }
          else {
            $fields_not_found[] = $field_name;
          }
        }
      }

      // Build comprehensive token data structure.
      $token_data = [
        // Success indicator.
        'success' => 1,
        'error' => '',
        
        // Object metadata.
        'object_type' => $sf_object->type(),
        'record_id' => (string) $sf_object->id(),
        
        // Field data.
        'fields' => $field_values,
        'field_count' => count($field_values),
        'all_field_names' => array_keys($all_fields),
        'requested_fields' => $requested_fields,
        'fields_not_found' => $fields_not_found,
        
        // Direct field access (for convenience).
        // Each field is also available as a direct token.
      ];

      // Add individual field tokens for direct access.
      foreach ($field_values as $field_name => $field_value) {
        $token_data[$field_name] = $field_value;
      }

      // Store the token data.
      $this->tokenService->addTokenData($result_token, $token_data);

      // Log success.
      $this->logger->info('Successfully retrieved @count field(s) from Salesforce @type/@id', [
        '@count' => count($field_values),
        '@type' => $object_type,
        '@id' => $sf_object->id(),
      ]);

      // Log warning if some fields were not found.
      if (!empty($fields_not_found)) {
        $this->logger->warning('Fields not found in Salesforce object: @fields', [
          '@fields' => implode(', ', $fields_not_found),
        ]);
      }

    }
    catch (\Exception $e) {
      // Log the error.
      $this->logger->error('Error getting Salesforce field values: @error', [
        '@error' => $e->getMessage(),
      ]);

      // Populate error token data.
      $this->tokenService->addTokenData($result_token, [
        'success' => 0,
        'error' => $e->getMessage(),
        'object_type' => $object_type ?? '',
        'record_id' => $record_id ?? '',
        'fields' => [],
        'field_count' => 0,
        'all_field_names' => [],
        'requested_fields' => $requested_fields ?? [],
        'fields_not_found' => [],
      ]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'object_type' => '',
      'record_id' => '',
      'use_external_id' => FALSE,
      'external_id_field' => '',
      'external_id_value' => '',
      'field_names' => '',
      'result_token' => 'salesforce_field_values',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['object_type'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Salesforce Object Type'),
      '#description' => $this->t('The Salesforce object type (e.g., Contact, Account, Lead, CustomObject__c). You can use tokens.'),
      '#default_value' => $this->configuration['object_type'],
      '#required' => TRUE,
    ];

    // ID method selection.
    $form['use_external_id'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use External ID'),
      '#description' => $this->t('Check this to use an external ID instead of the Salesforce record ID.'),
      '#default_value' => $this->configuration['use_external_id'],
    ];

    // Standard Salesforce ID.
    $form['record_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Salesforce Record ID'),
      '#description' => $this->t('The Salesforce record ID (e.g., 003XX000004TMM2AAO). You can use tokens.'),
      '#default_value' => $this->configuration['record_id'],
      '#states' => [
        'visible' => [
          ':input[name="use_external_id"]' => ['checked' => FALSE],
        ],
        'required' => [
          ':input[name="use_external_id"]' => ['checked' => FALSE],
        ],
      ],
    ];

    // External ID fields.
    $form['external_id_field'] = [
      '#type' => 'textfield',
      '#title' => $this->t('External ID Field Name'),
      '#description' => $this->t('The name of the external ID field in Salesforce (e.g., External_ID__c, Email). You can use tokens.'),
      '#default_value' => $this->configuration['external_id_field'],
      '#states' => [
        'visible' => [
          ':input[name="use_external_id"]' => ['checked' => TRUE],
        ],
        'required' => [
          ':input[name="use_external_id"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $form['external_id_value'] = [
      '#type' => 'textfield',
      '#title' => $this->t('External ID Value'),
      '#description' => $this->t('The value of the external ID field (e.g., user@example.com, 12345). You can use tokens.'),
      '#default_value' => $this->configuration['external_id_value'],
      '#states' => [
        'visible' => [
          ':input[name="use_external_id"]' => ['checked' => TRUE],
        ],
        'required' => [
          ':input[name="use_external_id"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $form['field_names'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Field Names'),
      '#description' => $this->t('Comma-separated list of field names to retrieve. Leave empty or use "*" to get all fields. You can use tokens.<br>
        <strong>Examples:</strong><br>
        • <code>Email,FirstName,LastName</code> - Get specific fields<br>
        • <code>Name,Account.Name,Owner.Email</code> - Include relationship fields<br>
        • <code>*</code> or empty - Get all fields<br>
        <strong>Note:</strong> Getting all fields uses objectRead() which is more efficient than SOQL for single records.'),
      '#default_value' => $this->configuration['field_names'],
      '#required' => FALSE,
      '#rows' => 3,
    ];

    $form['result_token'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Result token name'),
      '#description' => $this->t('The token name to store the field values. Access data using:<br>
        <strong>Status tokens:</strong><br>
        • <code>[TOKEN:success]</code> - Whether the operation succeeded (1 or 0)<br>
        • <code>[TOKEN:error]</code> - Error message if failed<br>
        <strong>Object metadata:</strong><br>
        • <code>[TOKEN:object_type]</code> - The Salesforce object type<br>
        • <code>[TOKEN:record_id]</code> - The Salesforce record ID<br>
        <strong>Field data:</strong><br>
        • <code>[TOKEN:fields]</code> - Array of all retrieved fields<br>
        • <code>[TOKEN:field_count]</code> - Number of fields retrieved<br>
        • <code>[TOKEN:all_field_names]</code> - All available field names<br>
        • <code>[TOKEN:requested_fields]</code> - Fields that were requested<br>
        • <code>[TOKEN:fields_not_found]</code> - Requested fields that don\'t exist<br>
        <strong>Direct field access:</strong><br>
        • <code>[TOKEN:Email]</code> - Direct access to Email field<br>
        • <code>[TOKEN:FirstName]</code> - Direct access to FirstName field<br>
        • <code>[TOKEN:CustomField__c]</code> - Direct access to custom field<br>
        <strong>Note:</strong> Field names are case-sensitive and must match Salesforce exactly.'),
      '#default_value' => $this->configuration['result_token'],
      '#required' => TRUE,
      '#eca_token_reference' => TRUE,
      '#element_validate' => [[$this, 'validateTokenName']],
    ];

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * Validates the token name field.
   *
   * @param array $element
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function validateTokenName(array $element, FormStateInterface $form_state): void {
    $value = $element['#value'];
    if (!preg_match('/^[a-zA-Z_][a-zA-Z0-9_]*$/', $value)) {
      $form_state->setError($element, $this->t('Token name must start with a letter or underscore and contain only letters, numbers, and underscores.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->configuration['object_type'] = $form_state->getValue('object_type');
    $this->configuration['record_id'] = $form_state->getValue('record_id');
    $this->configuration['use_external_id'] = (bool) $form_state->getValue('use_external_id');
    $this->configuration['external_id_field'] = $form_state->getValue('external_id_field');
    $this->configuration['external_id_value'] = $form_state->getValue('external_id_value');
    $this->configuration['field_names'] = $form_state->getValue('field_names');
    $this->configuration['result_token'] = $form_state->getValue('result_token');
    parent::submitConfigurationForm($form, $form_state);
  }

}
