<?php

namespace Drupal\salesforce_eca\Plugin\Action;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\eca\Plugin\Action\ConfigurableActionBase;
use Drupal\salesforce_mapping\MappedObjectStorage;
use Drupal\salesforce_mapping\SalesforceMappingStorage;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Get Salesforce mapped object information for an entity.
 *
 * @Action(
 *   id = "salesforce_get_mapped_object",
 *   label = @Translation("Get Salesforce mapped object"),
 *   description = @Translation("Load the Salesforce mapped object information for a Drupal entity. Provides the Salesforce ID and mapping details needed for other actions."),
 *   eca_version_introduced = "1.0.0",
 *   type = "entity"
 * )
 */
class SalesforceGetMappedObject extends ConfigurableActionBase {

  /**
   * The Salesforce mapped object storage.
   *
   * @var \Drupal\salesforce_mapping\MappedObjectStorage
   */
  protected MappedObjectStorage $mappedObjectStorage;

  /**
   * The Salesforce mapping storage.
   *
   * @var \Drupal\salesforce_mapping\SalesforceMappingStorage
   */
  protected SalesforceMappingStorage $mappingStorage;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('eca.token_services'),
      $container->get('current_user'),
      $container->get('datetime.time'),
      $container->get('eca.state'),
      $container->get('logger.channel.eca')
    );

    // Inject the additional services.
    $instance->setMappedObjectStorage($container->get('entity_type.manager')->getStorage('salesforce_mapped_object'));
    $instance->setMappingStorage($container->get('entity_type.manager')->getStorage('salesforce_mapping'));

    return $instance;
  }

  /**
   * Sets the mapped object storage.
   *
   * @param \Drupal\salesforce_mapping\MappedObjectStorage $mapped_object_storage
   *   The mapped object storage.
   */
  public function setMappedObjectStorage(MappedObjectStorage $mapped_object_storage): void {
    $this->mappedObjectStorage = $mapped_object_storage;
  }

  /**
   * Sets the mapping storage.
   *
   * @param \Drupal\salesforce_mapping\SalesforceMappingStorage $mapping_storage
   *   The mapping storage.
   */
  public function setMappingStorage(SalesforceMappingStorage $mapping_storage): void {
    $this->mappingStorage = $mapping_storage;
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, ?AccountInterface $account = NULL, $return_as_object = FALSE): bool|AccessResultInterface {
    // ECA actions generally allow access - workflow controls execution.
    $access_result = AccessResult::allowed();
    return $return_as_object ? $access_result : $access_result->isAllowed();
  }

  /**
   * {@inheritdoc}
   */
  public function execute($entity = NULL): void {
    $result_token = $this->configuration['result_token'];

    try {
      // Validate entity.
      if (!$entity instanceof EntityInterface) {
        throw new \Exception('No valid entity provided.');
      }

      $this->logger->info('Getting Salesforce mapped object for @type/@id', [
        '@type' => $entity->getEntityTypeId(),
        '@id' => $entity->id(),
      ]);

      // Load mapped objects for this entity.
      $mapped_objects = $this->mappedObjectStorage->loadByEntity($entity);

      if (empty($mapped_objects)) {
        // No mapped objects found.
        $this->logger->info('No Salesforce mapped objects found for @type/@id', [
          '@type' => $entity->getEntityTypeId(),
          '@id' => $entity->id(),
        ]);

        $this->tokenService->addTokenData($result_token, [
          'success' => 0,
          'error' => 'No Salesforce mapping found for this entity',
          'is_mapped' => 0,
          'mapping_count' => 0,
          'salesforce_id' => '',
          'salesforce_object_type' => '',
          'mapping_id' => '',
          'last_sync_status' => '',
          'last_sync_action' => '',
          'entity_updated' => '',
          'mapped_object_id' => '',
        ]);
        return;
      }

      // Check if specific mapping requested.
      $specific_mapping_id = $this->configuration['mapping_id'];
      if ($specific_mapping_id) {
        $specific_mapping_id = $this->tokenService->replaceClear($specific_mapping_id);
        // Filter to specific mapping.
        $mapped_objects = array_filter($mapped_objects, function ($mapped_object) use ($specific_mapping_id) {
          return $mapped_object->get('salesforce_mapping')->target_id === $specific_mapping_id;
        });

        if (empty($mapped_objects)) {
          throw new \Exception('Specified mapping not found for this entity.');
        }
      }

      // Get the first (or only) mapped object.
      $mapped_object = reset($mapped_objects);
      
      // Load the mapping to get object type.
      $mapping = $mapped_object->getMapping();
      $salesforce_object_type = $mapping ? $mapping->getSalesforceObjectType() : '';

      // Build comprehensive token data.
      $token_data = [
        // Success indicators.
        'success' => 1,
        'error' => '',
        'is_mapped' => 1,
        'mapping_count' => count($mapped_objects),
        
        // Primary mapped object data.
        'salesforce_id' => $mapped_object->sfid(),
        'salesforce_object_type' => $salesforce_object_type,
        'mapping_id' => $mapped_object->get('salesforce_mapping')->target_id,
        'mapped_object_id' => $mapped_object->id(),
        
        // Sync status information.
        'last_sync_status' => $mapped_object->get('last_sync_status')->value ? 1 : 0,
        'last_sync_action' => $mapped_object->get('last_sync_action')->value ?? '',
        'entity_updated' => $mapped_object->get('entity_updated')->value ?? '',
        'force_pull' => $mapped_object->get('force_pull')->value ? 1 : 0,
        
        // Salesforce URL.
        'salesforce_url' => $mapped_object->getSalesforceUrl(),
      ];

      // If multiple mappings exist, provide info about all.
      if (count($mapped_objects) > 1) {
        $all_mappings = [];
        foreach ($mapped_objects as $mo) {
          $mapping = $mo->getMapping();
          $all_mappings[] = [
            'salesforce_id' => $mo->sfid(),
            'mapping_id' => $mo->get('salesforce_mapping')->target_id,
            'object_type' => $mapping ? $mapping->getSalesforceObjectType() : '',
          ];
        }
        $token_data['all_mappings'] = $all_mappings;
      }

      // Store the token data.
      $this->tokenService->addTokenData($result_token, $token_data);

      // Log success.
      $this->logger->info('Found @count Salesforce mapping(s) for @type/@id. Primary: @sfid', [
        '@count' => count($mapped_objects),
        '@type' => $entity->getEntityTypeId(),
        '@id' => $entity->id(),
        '@sfid' => $mapped_object->sfid(),
      ]);

    }
    catch (\Exception $e) {
      // Log the error.
      $this->logger->error('Error getting Salesforce mapped object: @error', [
        '@error' => $e->getMessage(),
      ]);

      // Populate error token data.
      $this->tokenService->addTokenData($result_token, [
        'success' => 0,
        'error' => $e->getMessage(),
        'is_mapped' => 0,
        'mapping_count' => 0,
        'salesforce_id' => '',
        'salesforce_object_type' => '',
        'mapping_id' => '',
        'last_sync_status' => '',
        'last_sync_action' => '',
        'entity_updated' => '',
        'mapped_object_id' => '',
      ]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'mapping_id' => '',
      'result_token' => 'salesforce_mapped_object',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['mapping_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Salesforce Mapping ID'),
      '#description' => $this->t('Optional: Specify a Salesforce mapping ID to retrieve. If left empty, the first mapped object found will be used. You can use tokens in this field.'),
      '#default_value' => $this->configuration['mapping_id'],
      '#required' => FALSE,
    ];

    $form['result_token'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Result token name'),
      '#description' => $this->t('The token name to store the mapped object information. Access data using:<br>
        <strong>Mapping status:</strong><br>
        • <code>[TOKEN:success]</code> - Whether a mapping was found (1 or 0)<br>
        • <code>[TOKEN:is_mapped]</code> - Same as success<br>
        • <code>[TOKEN:mapping_count]</code> - Number of mappings found<br>
        • <code>[TOKEN:error]</code> - Error message if any<br>
        <strong>Salesforce data:</strong><br>
        • <code>[TOKEN:salesforce_id]</code> - The Salesforce record ID<br>
        • <code>[TOKEN:salesforce_object_type]</code> - Object type (Contact, Lead, etc.)<br>
        • <code>[TOKEN:mapping_id]</code> - The Salesforce mapping configuration ID<br>
        • <code>[TOKEN:salesforce_url]</code> - Direct URL to Salesforce record<br>
        <strong>Sync information:</strong><br>
        • <code>[TOKEN:last_sync_status]</code> - Whether last sync succeeded (1 or 0)<br>
        • <code>[TOKEN:last_sync_action]</code> - Last action (push_create, pull_update, etc.)<br>
        • <code>[TOKEN:entity_updated]</code> - Timestamp of last entity update<br>
        • <code>[TOKEN:force_pull]</code> - Whether force pull is flagged<br>
        <strong>Multiple mappings:</strong><br>
        • <code>[TOKEN:all_mappings]</code> - Array of all mappings if multiple exist'),
      '#default_value' => $this->configuration['result_token'],
      '#required' => TRUE,
      '#eca_token_reference' => TRUE,
      '#element_validate' => [[$this, 'validateTokenName']],
    ];

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * Validates the token name field.
   *
   * @param array $element
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function validateTokenName(array $element, FormStateInterface $form_state): void {
    $value = $element['#value'];
    if (!preg_match('/^[a-zA-Z_][a-zA-Z0-9_]*$/', $value)) {
      $form_state->setError($element, $this->t('Token name must start with a letter or underscore and contain only letters, numbers, and underscores.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->configuration['mapping_id'] = $form_state->getValue('mapping_id');
    $this->configuration['result_token'] = $form_state->getValue('result_token');
    parent::submitConfigurationForm($form, $form_state);
  }

}
