<?php

namespace Drupal\salesforce_eca\Plugin\Action;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\eca\Plugin\Action\ConfigurableActionBase;
use Drupal\salesforce\Rest\RestClientInterface;
use Drupal\salesforce\SFID;
use Drupal\salesforce_mapping\MappedObjectStorage;
use Drupal\salesforce_mapping\SalesforceMappingStorage;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Trigger an immediate pull of Salesforce data to an entity.
 *
 * @Action(
 *   id = "salesforce_trigger_pull",
 *   label = @Translation("Trigger entity pull from Salesforce"),
 *   description = @Translation("Pull latest data from Salesforce for an entity using its configured mapping. The entity must already be mapped to a Salesforce record."),
 *   eca_version_introduced = "1.0.0",
 *   type = "entity"
 * )
 */
class SalesforceTriggerPull extends ConfigurableActionBase {

  /**
   * The Salesforce REST client.
   *
   * @var \Drupal\salesforce\Rest\RestClientInterface
   */
  protected RestClientInterface $salesforceClient;

  /**
   * The Salesforce mapped object storage.
   *
   * @var \Drupal\salesforce_mapping\MappedObjectStorage
   */
  protected MappedObjectStorage $mappedObjectStorage;

  /**
   * The Salesforce mapping storage.
   *
   * @var \Drupal\salesforce_mapping\SalesforceMappingStorage
   */
  protected SalesforceMappingStorage $mappingStorage;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('eca.token_services'),
      $container->get('current_user'),
      $container->get('datetime.time'),
      $container->get('eca.state'),
      $container->get('logger.channel.eca')
    );

    // Inject the additional services.
    $instance->setSalesforceClient($container->get('salesforce.client'));
    $instance->setMappedObjectStorage($container->get('entity_type.manager')->getStorage('salesforce_mapped_object'));
    $instance->setMappingStorage($container->get('entity_type.manager')->getStorage('salesforce_mapping'));

    return $instance;
  }

  /**
   * Sets the Salesforce client.
   *
   * @param \Drupal\salesforce\Rest\RestClientInterface $salesforce_client
   *   The Salesforce REST client.
   */
  public function setSalesforceClient(RestClientInterface $salesforce_client): void {
    $this->salesforceClient = $salesforce_client;
  }

  /**
   * Sets the mapped object storage.
   *
   * @param \Drupal\salesforce_mapping\MappedObjectStorage $mapped_object_storage
   *   The mapped object storage.
   */
  public function setMappedObjectStorage(MappedObjectStorage $mapped_object_storage): void {
    $this->mappedObjectStorage = $mapped_object_storage;
  }

  /**
   * Sets the mapping storage.
   *
   * @param \Drupal\salesforce_mapping\SalesforceMappingStorage $mapping_storage
   *   The mapping storage.
   */
  public function setMappingStorage(SalesforceMappingStorage $mapping_storage): void {
    $this->mappingStorage = $mapping_storage;
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, ?AccountInterface $account = NULL, $return_as_object = FALSE): bool|AccessResultInterface {
    // For ECA actions, we generally allow access since the action execution
    // is controlled by the ECA workflow configuration.
    $access_result = AccessResult::allowed();
    return $return_as_object ? $access_result : $access_result->isAllowed();
  }

  /**
   * {@inheritdoc}
   */
  public function execute($entity = NULL): void {
    $result_token = $this->configuration['result_token'];

    // Check if we should pull by Salesforce ID instead of entity.
    $salesforce_id = $this->configuration['salesforce_id'];
    if ($salesforce_id) {
      $salesforce_id = $this->tokenService->replaceClear($salesforce_id);

      // Find mapped object by Salesforce ID.
      try {
        $sfid = new SFID($salesforce_id);
      }
      catch (\Exception $e) {
        $this->logger->error('Invalid Salesforce ID format: @sfid', [
          '@sfid' => $salesforce_id,
        ]);
        $this->tokenService->addTokenData($result_token, [
          'success' => 0,
          'error' => 'Invalid Salesforce ID format',
          'entity_id' => '',
          'entity_type' => '',
          'updated_fields' => [],
        ]);
        return;
      }

      $mapped_objects = $this->mappedObjectStorage->loadBySfid($sfid);

      if (empty($mapped_objects)) {
        $this->logger->error('No mapped object found for Salesforce ID @sfid.', [
          '@sfid' => $salesforce_id,
        ]);
        $this->tokenService->addTokenData($result_token, [
          'success' => 0,
          'error' => 'No mapped object found for Salesforce ID',
          'entity_id' => '',
          'entity_type' => '',
          'updated_fields' => [],
        ]);
        return;
      }

      // Use the first mapped object found.
      $mapped_object = reset($mapped_objects);
      $entity = $mapped_object->getMappedEntity();
    }
    elseif ($entity instanceof EntityInterface) {
      // Load mapped objects for the entity.
      $mapped_objects = $this->mappedObjectStorage->loadByEntity($entity);

      if (empty($mapped_objects)) {
        $this->logger->error('No Salesforce mapped objects found for @type/@id.', [
          '@type' => $entity->getEntityTypeId(),
          '@id' => $entity->id(),
        ]);
        $this->tokenService->addTokenData($result_token, [
          'success' => 0,
          'error' => 'Entity is not mapped to Salesforce',
          'entity_id' => $entity->id(),
          'entity_type' => $entity->getEntityTypeId(),
          'updated_fields' => [],
        ]);
        return;
      }

      // Check if we should pull only specific mapping.
      $specific_mapping_id = $this->configuration['mapping_id'];
      if ($specific_mapping_id) {
        $specific_mapping_id = $this->tokenService->replaceClear($specific_mapping_id);
        // Filter mapped objects to only the specified mapping.
        $mapped_objects = array_filter($mapped_objects, function ($mapped_object) use ($specific_mapping_id) {
          return $mapped_object->getMapping()->id() === $specific_mapping_id;
        });

        if (empty($mapped_objects)) {
          $this->logger->warning('No mapped object found for specified mapping @id.', [
            '@id' => $specific_mapping_id,
          ]);
          $this->tokenService->addTokenData($result_token, [
            'success' => 0,
            'error' => 'No mapped object found for specified mapping',
            'entity_id' => $entity->id(),
            'entity_type' => $entity->getEntityTypeId(),
            'updated_fields' => [],
          ]);
          return;
        }
      }
    }
    else {
      $this->logger->error('Salesforce trigger pull: No valid entity or Salesforce ID provided.');
      $this->tokenService->addTokenData($result_token, [
        'success' => 0,
        'error' => 'No valid entity or Salesforce ID provided',
        'entity_id' => '',
        'entity_type' => '',
        'updated_fields' => [],
      ]);
      return;
    }

    try {
      // Process each mapped object.
      $success_count = 0;
      $error_messages = [];
      $all_updated_fields = [];

      foreach ($mapped_objects as $mapped_object) {
        try {
          // Get the entity before pull to track changes.
          $entity_before = clone $mapped_object->getMappedEntity();

          // Perform the pull operation.
          $mapped_object->pull();

          // Pull succeeded if no exception was thrown.
          $success_count++;

          // Get the entity after pull.
          $entity_after = $mapped_object->getMappedEntity();

          // Track which fields were updated.
          $updated_fields = [];
          if ($entity_before && $entity_after) {
            foreach ($entity_after->getFields() as $field_name => $field) {
              if ($entity_before->hasField($field_name)) {
                $before_value = $entity_before->get($field_name)->getValue();
                $after_value = $entity_after->get($field_name)->getValue();

                if (serialize($before_value) !== serialize($after_value)) {
                  $updated_fields[] = $field_name;
                }
              }
            }
          }

          $all_updated_fields = array_unique(array_merge($all_updated_fields, $updated_fields));

          $this->logger->info('Successfully pulled Salesforce data for @type/@id from @sfid using mapping @mapping.', [
            '@type' => $entity->getEntityTypeId(),
            '@id' => $entity->id(),
            '@sfid' => $mapped_object->sfid(),
            '@mapping' => $mapped_object->getMapping()->id(),
          ]);
        }
        catch (\Exception $e) {
          $error_messages[] = "Pull error for mapping {$mapped_object->getMapping()->id()}: " . $e->getMessage();
          $this->logger->error('Error pulling entity from Salesforce: @error', [
            '@error' => $e->getMessage(),
          ]);
        }
      }

      // Store results in token.
      $this->tokenService->addTokenData($result_token, [
        'success' => $success_count > 0 ? 1 : 0,
        'success_count' => $success_count,
        'total_mappings' => count($mapped_objects),
        'error' => implode('; ', $error_messages),
        'entity_id' => $entity ? $entity->id() : '',
        'entity_type' => $entity ? $entity->getEntityTypeId() : '',
        'updated_fields' => $all_updated_fields,
        'updated_fields_count' => count($all_updated_fields),
      ]);

    }
    catch (\Exception $e) {
      $this->logger->error('Unexpected error during Salesforce pull: @error', [
        '@error' => $e->getMessage(),
      ]);
      $this->tokenService->addTokenData($result_token, [
        'success' => 0,
        'error' => 'Unexpected error: ' . $e->getMessage(),
        'entity_id' => $entity ? $entity->id() : '',
        'entity_type' => $entity ? $entity->getEntityTypeId() : '',
        'updated_fields' => [],
      ]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'salesforce_id' => '',
      'mapping_id' => '',
      'result_token' => 'salesforce_pull_result',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['salesforce_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Salesforce ID'),
      '#description' => $this->t('Optional: Pull data for a specific Salesforce ID. If provided, this takes precedence over the entity context. You can use tokens in this field.'),
      '#default_value' => $this->configuration['salesforce_id'],
      '#required' => FALSE,
    ];

    $form['mapping_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Salesforce Mapping ID'),
      '#description' => $this->t('Optional: Specify a Salesforce mapping ID to use for the pull. If left empty, all applicable pull mappings for the entity will be processed. You can use tokens in this field.'),
      '#default_value' => $this->configuration['mapping_id'],
      '#required' => FALSE,
    ];

    $form['result_token'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Result token name'),
      '#description' => $this->t('The token name to store the pull operation results. The token will contain:<br>
        • <code>[TOKEN:success]</code> - Whether the pull succeeded (1 or 0)<br>
        • <code>[TOKEN:success_count]</code> - Number of successful pulls<br>
        • <code>[TOKEN:total_mappings]</code> - Total number of mappings processed<br>
        • <code>[TOKEN:error]</code> - Error message if pull failed<br>
        • <code>[TOKEN:entity_id]</code> - The ID of the updated entity<br>
        • <code>[TOKEN:entity_type]</code> - The type of the updated entity<br>
        • <code>[TOKEN:updated_fields]</code> - Array of field names that were updated<br>
        • <code>[TOKEN:updated_fields_count]</code> - Number of fields that were updated'),
      '#default_value' => $this->configuration['result_token'],
      '#required' => TRUE,
      '#eca_token_reference' => TRUE,
      '#element_validate' => [[$this, 'validateTokenName']],
    ];

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * Validates the token name field.
   *
   * @param array $element
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function validateTokenName(array $element, FormStateInterface $form_state): void {
    $value = $element['#value'];
    if (!preg_match('/^[a-zA-Z_][a-zA-Z0-9_]*$/', $value)) {
      $form_state->setError($element, $this->t('Token name must start with a letter or underscore and contain only letters, numbers, and underscores.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->configuration['salesforce_id'] = $form_state->getValue('salesforce_id');
    $this->configuration['mapping_id'] = $form_state->getValue('mapping_id');
    $this->configuration['result_token'] = $form_state->getValue('result_token');
    parent::submitConfigurationForm($form, $form_state);
  }

}
