<?php

namespace Drupal\salesforce_eca\Plugin\Action;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\eca\Plugin\Action\ConfigurableActionBase;
use Drupal\salesforce\Rest\RestClientInterface;
use Drupal\salesforce_mapping\MappedObjectStorage;
use Drupal\salesforce_mapping\SalesforceMappingStorage;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Trigger an immediate push of an entity to Salesforce.
 *
 * @Action(
 *   id = "salesforce_trigger_push",
 *   label = @Translation("Trigger entity push to Salesforce"),
 *   description = @Translation("Force immediate push of an entity to Salesforce using its configured mapping. The entity must have an active Salesforce mapping."),
 *   eca_version_introduced = "1.0.0",
 *   type = "entity"
 * )
 */
class SalesforceTriggerPush extends ConfigurableActionBase {

  /**
   * The Salesforce REST client.
   *
   * @var \Drupal\salesforce\Rest\RestClientInterface
   */
  protected RestClientInterface $salesforceClient;

  /**
   * The Salesforce mapped object storage.
   *
   * @var \Drupal\salesforce_mapping\MappedObjectStorage
   */
  protected MappedObjectStorage $mappedObjectStorage;

  /**
   * The Salesforce mapping storage.
   *
   * @var \Drupal\salesforce_mapping\SalesforceMappingStorage
   */
  protected SalesforceMappingStorage $mappingStorage;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('eca.token_services'),
      $container->get('current_user'),
      $container->get('datetime.time'),
      $container->get('eca.state'),
      $container->get('logger.channel.eca')
    );

    // Inject the additional services.
    $instance->setSalesforceClient($container->get('salesforce.client'));
    $instance->setMappedObjectStorage($container->get('entity_type.manager')->getStorage('salesforce_mapped_object'));
    $instance->setMappingStorage($container->get('entity_type.manager')->getStorage('salesforce_mapping'));

    return $instance;
  }

  /**
   * Sets the Salesforce client.
   *
   * @param \Drupal\salesforce\Rest\RestClientInterface $salesforce_client
   *   The Salesforce REST client.
   */
  public function setSalesforceClient(RestClientInterface $salesforce_client): void {
    $this->salesforceClient = $salesforce_client;
  }

  /**
   * Sets the mapped object storage.
   *
   * @param \Drupal\salesforce_mapping\MappedObjectStorage $mapped_object_storage
   *   The mapped object storage.
   */
  public function setMappedObjectStorage(MappedObjectStorage $mapped_object_storage): void {
    $this->mappedObjectStorage = $mapped_object_storage;
  }

  /**
   * Sets the mapping storage.
   *
   * @param \Drupal\salesforce_mapping\SalesforceMappingStorage $mapping_storage
   *   The mapping storage.
   */
  public function setMappingStorage(SalesforceMappingStorage $mapping_storage): void {
    $this->mappingStorage = $mapping_storage;
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, ?AccountInterface $account = NULL, $return_as_object = FALSE): bool|AccessResultInterface {
    // For ECA actions, we generally allow access since the action execution
    // is controlled by the ECA workflow configuration.
    $access_result = AccessResult::allowed();
    return $return_as_object ? $access_result : $access_result->isAllowed();
  }

  /**
   * {@inheritdoc}
   */
  public function execute($entity = NULL): void {
    try {
      $result_token = $this->configuration['result_token'];

      // Log that we're starting the push action.
      $this->logger->info('Salesforce trigger push: Starting push action for entity.');

      if (!$entity instanceof EntityInterface) {
        $this->logger->error('Salesforce trigger push: No valid entity provided.');
        $this->tokenService->addTokenData($result_token, [
          'success' => 0,
          'error' => 'No valid entity provided',
          'salesforce_id' => '',
          'mapped_object_id' => '',
        ]);
        return;
      }

      // Check if entity is already syncing to prevent recursion.
      if ($entity->isSyncing()) {
        $this->logger->info('Salesforce trigger push: Entity is already syncing, skipping.');
        $this->tokenService->addTokenData($result_token, [
          'success' => 1,
          'error' => '',
          'salesforce_id' => 'already_syncing',
          'mapped_object_id' => '',
        ]);
        return;
      }

      // Check if required services are available.
      if (!isset($this->mappingStorage) || !isset($this->mappedObjectStorage)) {
        $this->logger->error('Salesforce trigger push: Required services not available.');
        $this->tokenService->addTokenData($result_token, [
          'success' => 0,
          'error' => 'Required services not available',
          'salesforce_id' => '',
          'mapped_object_id' => '',
        ]);
        return;
      }

      $this->logger->info('Salesforce trigger push: Entity received: @type/@id', [
        '@type' => $entity->getEntityTypeId(),
        '@id' => $entity->id(),
      ]);

      try {
        // Check if entity has any Salesforce mappings.
        // First get all push mappings for this entity type.
        $this->logger->info('Salesforce trigger push: Loading push mappings for @type', [
          '@type' => $entity->getEntityTypeId(),
        ]);

        $all_mappings = $this->mappingStorage->loadPushMappings(
          $entity->getEntityTypeId()
        );

        $this->logger->info('Salesforce trigger push: Found @count total mappings', [
          '@count' => count($all_mappings),
        ]);

        // Then filter by bundle.
        $mappings = array_filter($all_mappings, function ($mapping) use ($entity) {
          return $mapping->get('drupal_bundle') === $entity->bundle();
        });

        $this->logger->info('Salesforce trigger push: @count mappings after bundle filter for @bundle', [
          '@count' => count($mappings),
          '@bundle' => $entity->bundle(),
        ]);

        if (empty($mappings)) {
          $this->logger->warning('No Salesforce push mappings found for @type/@bundle.', [
            '@type' => $entity->getEntityTypeId(),
            '@bundle' => $entity->bundle(),
          ]);
          $this->tokenService->addTokenData($result_token, [
            'success' => 0,
            'error' => 'No push mappings found for this entity type',
            'salesforce_id' => '',
            'mapped_object_id' => '',
          ]);
          return;
        }

        // Check if we should push only specific mapping.
        $specific_mapping_id = $this->configuration['mapping_id'];
        if ($specific_mapping_id) {
          $specific_mapping_id = $this->tokenService->replaceClear($specific_mapping_id);
          // Filter mappings to only the specified one.
          $mappings = array_filter($mappings, function ($mapping) use ($specific_mapping_id) {
            return $mapping->id() === $specific_mapping_id;
          });

          if (empty($mappings)) {
            $this->logger->warning('Specified Salesforce mapping @id not found or not applicable.', [
              '@id' => $specific_mapping_id,
            ]);
            $this->tokenService->addTokenData($result_token, [
              'success' => 0,
              'error' => 'Specified mapping not found or not applicable',
              'salesforce_id' => '',
              'mapped_object_id' => '',
            ]);
            return;
          }
        }

        // Process each applicable mapping.
        $success_count = 0;
        $error_messages = [];
        $salesforce_ids = [];
        $mapped_object_ids = [];

        foreach ($mappings as $mapping) {
          $this->logger->info('Salesforce trigger push: Processing mapping @id', [
            '@id' => $mapping->id(),
          ]);

          // Load or create the mapped object.
          $mapped_object = $this->mappedObjectStorage->loadByEntityAndMapping($entity, $mapping);

          if (!$mapped_object) {
            // Create a new mapped object if none exists.
            $mapped_object = $this->mappedObjectStorage->create([
              'drupal_entity' => $entity,
              'salesforce_mapping' => $mapping->id(),
            ]);
          }

          try {
            $this->logger->info('Salesforce trigger push: About to call push() on mapped object.');

            // Perform the push operation.
            $result = $mapped_object->push();

            $this->logger->info('Salesforce trigger push: push() returned: @result', [
              '@result' => var_export($result, TRUE),
            ]);

            // Push succeeded if no exception was thrown.
            $success_count++;
            $salesforce_ids[] = $mapped_object->sfid();
            $mapped_object_ids[] = $mapped_object->id();

            $this->logger->info('Successfully pushed @type/@id to Salesforce as @sfid using mapping @mapping.', [
              '@type' => $entity->getEntityTypeId(),
              '@id' => $entity->id(),
              '@sfid' => $mapped_object->sfid(),
              '@mapping' => $mapping->id(),
            ]);
          }
          catch (\Exception $e) {
            $error_messages[] = "Push error for mapping {$mapping->id()}: " . $e->getMessage();
            $this->logger->error('Error pushing entity to Salesforce: @error', [
              '@error' => $e->getMessage(),
            ]);
          }
        }

        // Store results in token.
        $this->tokenService->addTokenData($result_token, [
          'success' => $success_count > 0 ? 1 : 0,
          'success_count' => $success_count,
          'total_mappings' => count($mappings),
          'error' => implode('; ', $error_messages),
          'salesforce_id' => implode(',', $salesforce_ids),
          'salesforce_ids' => $salesforce_ids,
          'mapped_object_id' => implode(',', $mapped_object_ids),
          'mapped_object_ids' => $mapped_object_ids,
        ]);

        $this->logger->info('Salesforce trigger push: Completed. Success: @success/@total', [
          '@success' => $success_count,
          '@total' => count($mappings),
        ]);

      }
      catch (\Exception $e) {
        $this->logger->error('Unexpected error during Salesforce push: @error', [
          '@error' => $e->getMessage(),
        ]);
        $this->tokenService->addTokenData($result_token, [
          'success' => 0,
          'error' => 'Unexpected error: ' . $e->getMessage(),
          'salesforce_id' => '',
          'mapped_object_id' => '',
        ]);
      }
    }
    catch (\Throwable $t) {
      // Catch any throwable to prevent ECA workflow from breaking.
      $this->logger->error('Critical error in Salesforce trigger push: @error in @file:@line', [
        '@error' => $t->getMessage(),
        '@file' => $t->getFile(),
        '@line' => $t->getLine(),
      ]);
      if (isset($result_token)) {
        $this->tokenService->addTokenData($result_token, [
          'success' => 0,
          'error' => 'Critical error: ' . $t->getMessage(),
          'salesforce_id' => '',
          'mapped_object_id' => '',
        ]);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'mapping_id' => '',
      'result_token' => 'salesforce_push_result',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['mapping_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Salesforce Mapping ID'),
      '#description' => $this->t('Optional: Specify a Salesforce mapping ID to use for the push. If left empty, all applicable push mappings for the entity will be processed. You can use tokens in this field.'),
      '#default_value' => $this->configuration['mapping_id'],
      '#required' => FALSE,
    ];

    $form['result_token'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Result token name'),
      '#description' => $this->t('The token name to store the push operation results. The token will contain:<br>
        • <code>[TOKEN:success]</code> - Whether the push succeeded (1 or 0)<br>
        • <code>[TOKEN:success_count]</code> - Number of successful pushes<br>
        • <code>[TOKEN:total_mappings]</code> - Total number of mappings processed<br>
        • <code>[TOKEN:error]</code> - Error message if push failed<br>
        • <code>[TOKEN:salesforce_id]</code> - The Salesforce ID(s) of the pushed record(s)<br>
        • <code>[TOKEN:salesforce_ids]</code> - Array of all Salesforce IDs<br>
        • <code>[TOKEN:mapped_object_id]</code> - The mapped object ID(s)<br>
        • <code>[TOKEN:mapped_object_ids]</code> - Array of all mapped object IDs'),
      '#default_value' => $this->configuration['result_token'],
      '#required' => TRUE,
      '#eca_token_reference' => TRUE,
      '#element_validate' => [[$this, 'validateTokenName']],
    ];

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * Validates the token name field.
   *
   * @param array $element
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function validateTokenName(array $element, FormStateInterface $form_state): void {
    $value = $element['#value'];
    if (!preg_match('/^[a-zA-Z_][a-zA-Z0-9_]*$/', $value)) {
      $form_state->setError($element, $this->t('Token name must start with a letter or underscore and contain only letters, numbers, and underscores.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->configuration['mapping_id'] = $form_state->getValue('mapping_id');
    $this->configuration['result_token'] = $form_state->getValue('result_token');
    parent::submitConfigurationForm($form, $form_state);
  }

}
