<?php

namespace Drupal\salesforce_eca\Plugin\ECA\Condition;

use Drupal\Core\Annotation\ContextDefinition;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\eca\Plugin\ECA\Condition\ConditionBase;
use Drupal\salesforce_mapping\MappedObjectStorage;
use Drupal\salesforce_mapping\SalesforceMappingStorage;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Check if an entity has a Salesforce mapping.
 *
 * @EcaCondition(
 *   id = "salesforce_entity_is_mapped",
 *   label = @Translation("Entity is mapped to Salesforce"),
 *   description = @Translation("Checks if the entity has one or more Salesforce mapped objects. Optionally checks for a specific mapping."),
 *   eca_version_introduced = "1.0.0",
 *   context_definitions = {
 *     "entity" = @ContextDefinition("entity", label = @Translation("Entity"))
 *   }
 * )
 */
class EntityIsMapped extends ConditionBase {

  /**
   * The Salesforce mapped object storage.
   *
   * @var \Drupal\salesforce_mapping\MappedObjectStorage
   */
  protected MappedObjectStorage $mappedObjectStorage;

  /**
   * The Salesforce mapping storage.
   *
   * @var \Drupal\salesforce_mapping\SalesforceMappingStorage
   */
  protected SalesforceMappingStorage $mappingStorage;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    
    // Inject the additional services.
    $instance->mappedObjectStorage = $container->get('entity_type.manager')->getStorage('salesforce_mapped_object');
    $instance->mappingStorage = $container->get('entity_type.manager')->getStorage('salesforce_mapping');
    
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function evaluate(): bool {
    $entity = $this->getValueFromContext('entity');
    
    // If no entity context, check for entity from token.
    if (!$entity instanceof EntityInterface) {
      // Try to get entity from token if specified.
      $entity_token = $this->configuration['entity_token'];
      if ($entity_token) {
        $entity_token = $this->tokenService->replace($entity_token);
        if ($entity_token && is_object($entity_token) && $entity_token instanceof EntityInterface) {
          $entity = $entity_token;
        }
      }
    }
    
    if (!$entity instanceof EntityInterface) {
      return FALSE;
    }

    try {
      // Load mapped objects for this entity.
      $mapped_objects = $this->mappedObjectStorage->loadByEntity($entity);
      
      if (empty($mapped_objects)) {
        return FALSE;
      }
      
      // If specific mapping requested, check for it.
      $specific_mapping_id = $this->configuration['mapping_id'];
      if ($specific_mapping_id) {
        $specific_mapping_id = $this->tokenService->replaceClear($specific_mapping_id);
        
        foreach ($mapped_objects as $mapped_object) {
          if ($mapped_object->get('salesforce_mapping')->target_id === $specific_mapping_id) {
            // Store mapping info in state for other actions to use.
            $this->storeMatchedMapping($entity, $mapped_object);
            return TRUE;
          }
        }
        // Specific mapping not found.
        return FALSE;
      }
      
      // Any mapping exists - store the first one.
      $mapped_object = reset($mapped_objects);
      $this->storeMatchedMapping($entity, $mapped_object);
      
      return TRUE;
      
    }
    catch (\Exception $e) {
      // Logger may not be available in conditions, skip logging.
      // Ensure condition returns FALSE on error.
      return FALSE;
    }
  }

  /**
   * Store matched mapping information in tokens for use by subsequent actions.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The Drupal entity.
   * @param \Drupal\salesforce_mapping\Entity\MappedObjectInterface $mapped_object
   *   The mapped object.
   */
  protected function storeMatchedMapping(EntityInterface $entity, $mapped_object): void {
    $store_tokens = $this->configuration['store_mapping_tokens'];
    if (!$store_tokens) {
      return;
    }
    
    $token_name = $this->configuration['mapping_token_name'];
    if (!$token_name) {
      return;
    }
    
    try {
      $mapping = $mapped_object->getMapping();
      $salesforce_object_type = $mapping ? $mapping->getSalesforceObjectType() : '';
      
      // Store mapping data in token.
      $this->tokenService->addTokenData($token_name, [
        'salesforce_id' => $mapped_object->sfid(),
        'salesforce_object_type' => $salesforce_object_type,
        'mapping_id' => $mapped_object->get('salesforce_mapping')->target_id,
        'mapped_object_id' => $mapped_object->id(),
        'salesforce_url' => $mapped_object->getSalesforceUrl(),
        'last_sync_status' => $mapped_object->get('last_sync_status')->value ? 1 : 0,
        'last_sync_action' => $mapped_object->get('last_sync_action')->value ?? '',
      ]);
      
      // Logger may not be available in conditions, skip logging.
      
    }
    catch (\Exception $e) {
      // Logger may not be available in conditions, skip logging.
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'entity_token' => '',
      'mapping_id' => '',
      'store_mapping_tokens' => TRUE,
      'mapping_token_name' => 'salesforce_mapping',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form = parent::buildConfigurationForm($form, $form_state);
    
    $form['entity_token'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Entity token'),
      '#description' => $this->t('Optional: Specify a token that contains the entity to check. Leave empty to use the entity from context (e.g., from a "Content entity: View" event).'),
      '#default_value' => $this->configuration['entity_token'],
      '#required' => FALSE,
    ];
    
    $form['mapping_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Salesforce Mapping ID'),
      '#description' => $this->t('Optional: Check for a specific Salesforce mapping ID. Leave empty to check if ANY mapping exists. You can use tokens.'),
      '#default_value' => $this->configuration['mapping_id'],
      '#required' => FALSE,
    ];
    
    $form['store_mapping_tokens'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Store mapping information in tokens'),
      '#description' => $this->t('If checked, the matched mapping information will be stored in a token for use by subsequent actions.'),
      '#default_value' => $this->configuration['store_mapping_tokens'],
    ];
    
    $form['mapping_token_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Mapping token name'),
      '#description' => $this->t('The token name to store the mapping information. You can then use tokens like [TOKEN:salesforce_id] and [TOKEN:salesforce_object_type] in subsequent actions.'),
      '#default_value' => $this->configuration['mapping_token_name'],
      '#states' => [
        'visible' => [
          ':input[name="store_mapping_tokens"]' => ['checked' => TRUE],
        ],
        'required' => [
          ':input[name="store_mapping_tokens"]' => ['checked' => TRUE],
        ],
      ],
    ];
    
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->configuration['entity_token'] = $form_state->getValue('entity_token');
    $this->configuration['mapping_id'] = $form_state->getValue('mapping_id');
    $this->configuration['store_mapping_tokens'] = (bool) $form_state->getValue('store_mapping_tokens');
    $this->configuration['mapping_token_name'] = $form_state->getValue('mapping_token_name');
    parent::submitConfigurationForm($form, $form_state);
  }

}
