<?php

namespace Drupal\salesforce_eca\Plugin\ECA\Event;

use Drupal\Core\Entity\EntityInterface;
use Drupal\salesforce_mapping\PushParams;
use Drupal\eca\Entity\Objects\EcaEvent;
use Drupal\Core\Form\FormStateInterface;
use Drupal\eca\Attributes\Token;
use Drupal\eca\Event\Tag;
use Drupal\eca\Plugin\CleanupInterface;
use Drupal\eca\Plugin\ECA\Event\EventBase;
use Drupal\salesforce\Event\SalesforceEvents as SalesforceEventConstants;
use Drupal\salesforce\Event\SalesforceErrorEvent;
use Drupal\salesforce_mapping\Event\SalesforcePullEnqueueEvent;
use Drupal\salesforce_mapping\Event\SalesforcePullEntityValueEvent;
use Drupal\salesforce_mapping\Event\SalesforcePullEvent;
use Drupal\salesforce_mapping\Event\SalesforcePushAllowedEvent;
use Drupal\salesforce_mapping\Event\SalesforcePushOpEvent;
use Drupal\salesforce_mapping\Event\SalesforcePushParamsEvent;
use Drupal\salesforce_mapping\Event\SalesforceQueryEvent;
use Symfony\Contracts\EventDispatcher\Event;

/**
 * Plugin implementation of the ECA Events for Salesforce.
 *
 * @EcaEvent(
 *   id = "salesforce",
 *   deriver = "Drupal\salesforce_eca\Plugin\ECA\Event\SalesforceEventsDeriver",
 *   description = @Translation("Salesforce events for push, pull, and mapping operations."),
 *   eca_version_introduced = "1.0.0"
 * )
 */
class SalesforceEvents extends EventBase implements CleanupInterface {

  /**
   * {@inheritdoc}
   */
  public static function definitions(): array {
    $definitions = [];

    $definitions['push_allowed'] = [
      'label' => 'Salesforce: Push to Salesforce allowed',
      'description' => 'Before pushing an entity to Salesforce, allowing the push to be prevented.',
      'event_name' => SalesforceEventConstants::PUSH_ALLOWED,
      'event_class' => SalesforcePushAllowedEvent::class,
      'tags' => Tag::CONTENT | Tag::WRITE | Tag::RUNTIME | Tag::BEFORE,
    ];

    $definitions['push_mapping_object'] = [
      'label' => 'Salesforce: Push mapping object',
      'description' => 'Immediately before processing a push, allowing modification of the mapped object.',
      'event_name' => SalesforceEventConstants::PUSH_MAPPING_OBJECT,
      'event_class' => SalesforcePushOpEvent::class,
      'tags' => Tag::CONTENT | Tag::WRITE | Tag::RUNTIME | Tag::BEFORE,
    ];

    $definitions['push_params'] = [
      'label' => 'Salesforce: Push parameters',
      'description' => 'When preparing push parameters, allowing modification of data sent to Salesforce.',
      'event_name' => SalesforceEventConstants::PUSH_PARAMS,
      'event_class' => SalesforcePushParamsEvent::class,
      'tags' => Tag::CONTENT | Tag::WRITE | Tag::RUNTIME | Tag::BEFORE,
    ];

    $definitions['push_success'] = [
      'label' => 'Salesforce: Push success',
      'description' => 'After successfully pushing an entity to Salesforce.',
      'event_name' => SalesforceEventConstants::PUSH_SUCCESS,
      'event_class' => SalesforcePushParamsEvent::class,
      'tags' => Tag::CONTENT | Tag::WRITE | Tag::RUNTIME | Tag::AFTER,
    ];

    $definitions['push_fail'] = [
      'label' => 'Salesforce: Push failure',
      'description' => 'When pushing an entity to Salesforce fails.',
      'event_name' => SalesforceEventConstants::PUSH_FAIL,
      'event_class' => SalesforcePushOpEvent::class,
      'tags' => Tag::CONTENT | Tag::WRITE | Tag::RUNTIME | Tag::AFTER,
    ];

    $definitions['pull_query'] = [
      'label' => 'Salesforce: Pull query',
      'description' => 'Before querying Salesforce, allowing modification of the SOQL query.',
      'event_name' => SalesforceEventConstants::PULL_QUERY,
      'event_class' => SalesforceQueryEvent::class,
      'tags' => Tag::CONTENT | Tag::READ | Tag::RUNTIME | Tag::BEFORE,
    ];

    $definitions['pull_enqueue'] = [
      'label' => 'Salesforce: Pull enqueue',
      'description' => 'Before enqueueing records for pull, allowing filtering of which records to process.',
      'event_name' => SalesforceEventConstants::PULL_ENQUEUE,
      'event_class' => SalesforcePullEnqueueEvent::class,
      'tags' => Tag::CONTENT | Tag::READ | Tag::RUNTIME | Tag::BEFORE,
    ];

    $definitions['pull_prepull'] = [
      'label' => 'Salesforce: Pull from Salesforce allowed',
      'description' => 'Before pulling data, allowing the pull to be prevented.',
      'event_name' => SalesforceEventConstants::PULL_PREPULL,
      'event_class' => SalesforcePullEvent::class,
      'tags' => Tag::CONTENT | Tag::READ | Tag::RUNTIME | Tag::BEFORE,
    ];

    $definitions['pull_entity_value'] = [
      'label' => 'Salesforce: Pull entity value',
      'description' => 'Before assigning each field value during pull, allowing value modification.',
      'event_name' => SalesforceEventConstants::PULL_ENTITY_VALUE,
      'event_class' => SalesforcePullEntityValueEvent::class,
      'tags' => Tag::CONTENT | Tag::READ | Tag::RUNTIME | Tag::BEFORE,
    ];

    $definitions['pull_presave'] = [
      'label' => 'Salesforce: Pull presave',
      'description' => 'Before saving a pulled entity, allowing final modifications.',
      'event_name' => SalesforceEventConstants::PULL_PRESAVE,
      'event_class' => SalesforcePullEvent::class,
      'tags' => Tag::CONTENT | Tag::READ | Tag::RUNTIME | Tag::BEFORE,
    ];

    $definitions['error'] = [
      'label' => 'Salesforce: Error',
      'description' => 'When a Salesforce error occurs during operations.',
      'event_name' => SalesforceEventConstants::ERROR,
      'event_class' => SalesforceErrorEvent::class,
      'tags' => Tag::RUNTIME | Tag::AFTER,
    ];

    return $definitions;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'mapping_id' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    if ($this->eventClass() === SalesforcePushAllowedEvent::class ||
        $this->eventClass() === SalesforcePushParamsEvent::class ||
        $this->eventClass() === SalesforcePushOpEvent::class ||
        $this->eventClass() === SalesforceQueryEvent::class ||
        $this->eventClass() === SalesforcePullEvent::class ||
        $this->eventClass() === SalesforcePullEnqueueEvent::class ||
        $this->eventClass() === SalesforcePullEntityValueEvent::class) {
      // Note: SalesforceErrorEvent is excluded - error events are system-wide

      $form['mapping_id'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Salesforce mapping ID'),
        '#description' => $this->t('Optionally filter by Salesforce mapping ID. Leave empty to trigger for all mappings. You can find mapping IDs at /admin/config/salesforce/mappings.'),
        '#default_value' => $this->configuration['mapping_id'],
        '#required' => FALSE,
        '#placeholder' => $this->t('e.g., contact_mapping'),
        '#weight' => -10,
      ];
    }

    // Call parent to render token documentation.
    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    parent::submitConfigurationForm($form, $form_state);
    if ($form_state->hasValue('mapping_id')) {
      $this->configuration['mapping_id'] = $form_state->getValue('mapping_id');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function generateWildcard(string $eca_config_id, EcaEvent $ecaEvent): string {
    /** @var \Drupal\eca\Plugin\ECA\Event\EventBase $plugin */
    $plugin = $ecaEvent->getPlugin();
    switch ($plugin->getDerivativeId()) {
      case 'push_allowed':
      case 'push_mapping_object':
      case 'push_params':
      case 'push_success':
      case 'push_fail':
      case 'pull_query':
      case 'pull_enqueue':
      case 'pull_prepull':
      case 'pull_entity_value':
      case 'pull_presave':
        return $this->configuration['mapping_id'] ?: '*';

      case 'error':
        // Error events are system-wide, no filtering
        return '*';

      default:
        return '*';
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function appliesForWildcard(Event $event, string $event_name, string $wildcard): bool {
    // ERROR events always apply (no mapping filtering)
    if ($event instanceof SalesforceErrorEvent) {
      return TRUE;
    }

    if ($event instanceof SalesforcePushAllowedEvent ||
        $event instanceof SalesforcePushParamsEvent ||
        $event instanceof SalesforcePushOpEvent ||
        $event instanceof SalesforceQueryEvent ||
        $event instanceof SalesforcePullEvent ||
        $event instanceof SalesforcePullEnqueueEvent ||
        $event instanceof SalesforcePullEntityValueEvent) {

      if ($wildcard === '*') {
        return TRUE;
      }

      // Check if the mapping ID matches the wildcard.
      $mapping = $event->getMapping();
      if ($mapping && $mapping->id() === $wildcard) {
        return TRUE;
      }
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function cleanupAfterSuccessors(): void {
    switch ($this->getDerivativeId()) {
      case 'push_allowed':
      case 'push_mapping_object':
      case 'push_params':
      case 'push_success':
      case 'push_fail':
      case 'pull_query':
      case 'pull_enqueue':
      case 'pull_prepull':
      case 'pull_entity_value':
      case 'pull_presave':
      case 'error':
        // No specific cleanup needed for Salesforce events.
        break;
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function buildEventData(): array {
    $event = $this->event;
    $data = [];

    // Handle ERROR events separately
    if ($event instanceof SalesforceErrorEvent) {
      // Get all data from the error event - only what it actually provides
      $data += [
        'exception' => $event->getException(),
        'error_message' => $event->getMessage(),
        'error_context' => $event->getContext(),
        'log_level' => $event->getLevel(),
      ];

      // Parse context placeholders into individual tokens for convenience
      $context = $event->getContext();
      if (is_array($context)) {
        $data += [
          'entity_id' => $context['%id'] ?? NULL,
          'item' => $context['%item'] ?? NULL,
          'mapping_info' => $context['%mapping'] ?? NULL,
          'operation_type' => $context['%type'] ?? NULL,
          'failure_details' => $context['%fail'] ?? NULL,
        ];
      }
    }
    elseif ($event instanceof SalesforcePushAllowedEvent ||
        $event instanceof SalesforcePushParamsEvent ||
        $event instanceof SalesforcePushOpEvent ||
        $event instanceof SalesforceQueryEvent ||
        $event instanceof SalesforcePullEvent ||
        $event instanceof SalesforcePullEnqueueEvent ||
        $event instanceof SalesforcePullEntityValueEvent) {

      // Store entities directly (ECA handles these well)
      $data['mapping'] = $event->getMapping();
      $data['entity'] = method_exists($event, 'getEntity') ? $event->getEntity() : NULL;

      // Add simple properties for easy token access.
      $mapping = $event->getMapping();
      if ($mapping) {
        $data['mapping_id'] = $mapping->id();
        $data['mapping_label'] = $mapping->label();
        $data['mapping_salesforce_object_type'] = method_exists($mapping, 'getSalesforceObjectType') ? $mapping->getSalesforceObjectType() : NULL;
      }

      // Add mapped object data.
      if (method_exists($event, 'getMappedObject')) {
        $mappedObject = $event->getMappedObject();
        if ($mappedObject) {
          // Keep entity as-is.
          $data['mapped_object'] = $mappedObject;
          $data['mapped_object_id'] = method_exists($mappedObject, 'id') ? $mappedObject->id() : NULL;
          $data['mapped_object_salesforce_id'] = method_exists($mappedObject, 'sfid') ? $mappedObject->sfid() : NULL;
        }
      }

      // Add event-specific data.
      if ($event instanceof SalesforcePushAllowedEvent) {
        $data += [
          'push_allowed' => $event->isPushAllowed(),
        ];
      }
      elseif ($event instanceof SalesforcePushParamsEvent) {
        // Get the actual params array from the PushParams object.
        $pushParams = $event->getParams();
        $params = [];
        if ($pushParams instanceof PushParams) {
          $params = $pushParams->getParams();
        }
        elseif (is_array($pushParams)) {
          $params = $pushParams;
        }

        // Ensure all params are scalar or arrays of scalars.
        if (is_array($params)) {
          array_walk_recursive($params, function (&$value) {
            if (is_object($value) && method_exists($value, '__toString')) {
              $value = (string) $value;
            }
            elseif (is_object($value)) {
              // Convert objects to array if possible.
              $value = method_exists($value, 'toArray') ? $value->toArray() : (array) $value;
            }
          });
        }

        $data += [
          'mapped_object' => $event->getMappedObject(),
          'params' => $params,
        ];
      }
      elseif ($event instanceof SalesforcePushOpEvent) {
        $data += [
          'mapped_object' => $event->getMappedObject(),
          'operation' => $event->getOp(),
        ];
      }
      elseif ($event instanceof SalesforceQueryEvent) {
        // Convert query object to array for ECA compatibility.
        $query = $event->getQuery();
        $queryData = NULL;
        if ($query) {
          // The SelectQuery class has public properties we can access directly.
          $queryData = [
            'object_type' => $query->objectType ?? NULL,
            'fields' => $query->fields ?? [],
            'conditions' => $query->conditions ?? [],
            'order' => $query->order ?? [],
            'limit' => $query->limit ?? NULL,
            'offset' => $query->offset ?? NULL,
            // The full SOQL query string.
            'query_string' => (string) $query,
          ];
        }

        $data += [
          'query' => $queryData,
        ];
      }
      elseif ($event instanceof SalesforcePullEnqueueEvent) {
        // Convert Salesforce SObject to array for ECA compatibility.
        $record = $event->getRecord();
        $recordData = NULL;
        if ($record && method_exists($record, 'fields')) {
          $recordData = $record->fields();
        }

        // The getRecords() method returns a SelectQueryResult, not an
        // array of SObjects. We need to get the records from the result
        // object.
        $queryResult = $event->getRecords();
        $recordsData = [];
        if ($queryResult && method_exists($queryResult, 'records')) {
          $records = $queryResult->records();
          if (is_array($records)) {
            foreach ($records as $key => $rec) {
              if ($rec && method_exists($rec, 'fields')) {
                $recordsData[$key] = $rec->fields();
              }
            }
          }
        }

        $data += [
          'record' => $recordData,
          'records' => $recordsData,
          'force_pull' => $event->getForcePull(),
          'enqueue_allowed' => $event->isEnqueueAllowed(),
        ];
      }
      elseif ($event instanceof SalesforcePullEntityValueEvent) {
        // Convert field plugin to array if it's an object.
        $fieldPlugin = $event->getFieldPlugin();
        $fieldPluginData = NULL;
        if (is_object($fieldPlugin)) {
          // Get the plugin ID and configuration.
          $fieldPluginData = [
            'plugin_id' => method_exists($fieldPlugin, 'getPluginId') ? $fieldPlugin->getPluginId() : NULL,
            'configuration' => method_exists($fieldPlugin, 'getConfiguration') ? $fieldPlugin->getConfiguration() : [],
          ];
          // Also try getPluginDefinition for plugin ID.
          if (empty($fieldPluginData['plugin_id']) && method_exists($fieldPlugin, 'getPluginDefinition')) {
            $definition = $fieldPlugin->getPluginDefinition();
            $fieldPluginData['plugin_id'] = $definition['id'] ?? NULL;
          }
        }

        // The entity value might be complex, ensure it's scalar.
        $entityValue = $event->getEntityValue();
        if (is_object($entityValue) && method_exists($entityValue, '__toString')) {
          $entityValue = (string) $entityValue;
        }
        elseif (is_object($entityValue) && method_exists($entityValue, 'getValue')) {
          // TypedData objects.
          $entityValue = $entityValue->getValue();
        }
        elseif (is_object($entityValue) && !($entityValue instanceof EntityInterface)) {
          // Other objects - try to convert to array.
          $entityValue = method_exists($entityValue, 'toArray') ? $entityValue->toArray() : NULL;
        }

        $data += [
          'mapped_object' => $event->getMappedObject(),
          'field_plugin' => $fieldPluginData,
          'entity_value' => $entityValue,
        ];
      }
      elseif ($event instanceof SalesforcePullEvent) {
        $data += [
          'mapped_object' => $event->getMappedObject(),
          'pull_allowed' => $event->isPullAllowed(),
        ];
      }
    }

    $data += parent::buildEventData();

    // Final safety check: ensure all data values are ECA-compatible.
    array_walk_recursive($data, function (&$value) {
      if (is_object($value)) {
        // Entities are allowed.
        if ($value instanceof EntityInterface) {
          return;
        }
        // Try to convert to string.
        if (method_exists($value, '__toString')) {
          $value = (string) $value;
        }
        // Try to get value from TypedData.
        elseif (method_exists($value, 'getValue')) {
          $value = $value->getValue();
        }
        // Try to convert to array.
        elseif (method_exists($value, 'toArray')) {
          $value = $value->toArray();
        }
        // Last resort: cast to array.
        else {
          $value = (array) $value;
        }
      }
    });

    return $data;
  }

  /**
   * {@inheritdoc}
   */
  #[Token(
    name: 'exception',
    description: 'The exception object that caused the error (if available).',
    classes: [
      SalesforceErrorEvent::class,
    ],
  )]
  #[Token(
    name: 'error_message',
    description: 'The error message.',
    classes: [
      SalesforceErrorEvent::class,
    ],
  )]
  #[Token(
    name: 'error_context',
    description: 'The full context array with placeholder values.',
    classes: [
      SalesforceErrorEvent::class,
    ],
  )]
  #[Token(
    name: 'log_level',
    description: 'The error severity level (always "error").',
    classes: [
      SalesforceErrorEvent::class,
    ],
  )]
  #[Token(
    name: 'entity_id',
    description: 'The entity ID from error context.',
    classes: [
      SalesforceErrorEvent::class,
    ],
  )]
  #[Token(
    name: 'item',
    description: 'The item being processed from error context.',
    classes: [
      SalesforceErrorEvent::class,
    ],
  )]
  #[Token(
    name: 'mapping_info',
    description: 'The mapping information from error context.',
    classes: [
      SalesforceErrorEvent::class,
    ],
  )]
  #[Token(
    name: 'operation_type',
    description: 'The operation type from error context.',
    classes: [
      SalesforceErrorEvent::class,
    ],
  )]
  #[Token(
    name: 'failure_details',
    description: 'The failure details from error context.',
    classes: [
      SalesforceErrorEvent::class,
    ],
  )]
  #[Token(
    name: 'entity',
    description: 'The Drupal entity.',
    classes: [
      SalesforcePushAllowedEvent::class,
      SalesforcePushParamsEvent::class,
      SalesforcePushOpEvent::class,
      SalesforcePullEvent::class,
      SalesforcePullEntityValueEvent::class,
    ],
  )]
  #[Token(
    name: 'mapping',
    description: 'The Salesforce mapping configuration.',
    classes: [
      SalesforcePushAllowedEvent::class,
      SalesforcePushParamsEvent::class,
      SalesforcePushOpEvent::class,
      SalesforceQueryEvent::class,
      SalesforcePullEvent::class,
      SalesforcePullEnqueueEvent::class,
      SalesforcePullEntityValueEvent::class,
    ],
  )]
  #[Token(
    name: 'mapped_object',
    description: 'The Salesforce mapped object entity.',
    classes: [
      SalesforcePushParamsEvent::class,
      SalesforcePushOpEvent::class,
      SalesforcePullEvent::class,
      SalesforcePullEntityValueEvent::class,
    ],
  )]
  #[Token(
    name: 'mapping_id',
    description: 'The Salesforce mapping ID.',
    classes: [
      SalesforcePushAllowedEvent::class,
      SalesforcePushParamsEvent::class,
      SalesforcePushOpEvent::class,
      SalesforceQueryEvent::class,
      SalesforcePullEvent::class,
      SalesforcePullEnqueueEvent::class,
      SalesforcePullEntityValueEvent::class,
    ],
  )]
  #[Token(
    name: 'mapping_label',
    description: 'The Salesforce mapping label.',
    classes: [
      SalesforcePushAllowedEvent::class,
      SalesforcePushParamsEvent::class,
      SalesforcePushOpEvent::class,
      SalesforceQueryEvent::class,
      SalesforcePullEvent::class,
      SalesforcePullEnqueueEvent::class,
      SalesforcePullEntityValueEvent::class,
    ],
  )]
  #[Token(
    name: 'mapping_salesforce_object_type',
    description: 'The Salesforce object type.',
    classes: [
      SalesforcePushAllowedEvent::class,
      SalesforcePushParamsEvent::class,
      SalesforcePushOpEvent::class,
      SalesforceQueryEvent::class,
      SalesforcePullEvent::class,
      SalesforcePullEnqueueEvent::class,
      SalesforcePullEntityValueEvent::class,
    ],
  )]
  #[Token(
    name: 'mapped_object_id',
    description: 'The mapped object entity ID.',
    classes: [
      SalesforcePushParamsEvent::class,
      SalesforcePushOpEvent::class,
      SalesforcePullEvent::class,
      SalesforcePullEntityValueEvent::class,
    ],
  )]
  #[Token(
    name: 'mapped_object_salesforce_id',
    description: 'The Salesforce record ID.',
    classes: [
      SalesforcePushParamsEvent::class,
      SalesforcePushOpEvent::class,
      SalesforcePullEvent::class,
      SalesforcePullEntityValueEvent::class,
    ],
  )]
  #[Token(
    name: 'push_allowed',
    description: 'Whether the push to Salesforce is allowed.',
    classes: [
      SalesforcePushAllowedEvent::class,
    ],
  )]
  #[Token(
    name: 'pull_allowed',
    description: 'Whether the pull from Salesforce is allowed.',
    classes: [
      SalesforcePullEvent::class,
    ],
  )]
  #[Token(
    name: 'params',
    description: 'The parameters being sent to Salesforce.',
    classes: [
      SalesforcePushParamsEvent::class,
    ],
  )]
  #[Token(
    name: 'query',
    description: 'The SOQL query being executed.',
    classes: [
      SalesforceQueryEvent::class,
    ],
  )]
  #[Token(
    name: 'query_string',
    description: 'The full SOQL query string.',
    classes: [
      SalesforceQueryEvent::class,
    ],
  )]
  #[Token(
    name: 'record',
    description: 'The individual Salesforce record.',
    classes: [
      SalesforcePullEnqueueEvent::class,
    ],
  )]
  #[Token(
    name: 'records',
    description: 'All Salesforce records being processed.',
    classes: [
      SalesforcePullEnqueueEvent::class,
    ],
  )]
  #[Token(
    name: 'field_plugin',
    description: 'The field mapping plugin.',
    classes: [
      SalesforcePullEntityValueEvent::class,
    ],
  )]
  #[Token(
    name: 'entity_value',
    description: 'The field value being set.',
    classes: [
      SalesforcePullEntityValueEvent::class,
    ],
  )]
  #[Token(
    name: 'force_pull',
    description: 'Whether this is a forced pull operation.',
    classes: [
      SalesforcePullEnqueueEvent::class,
    ],
  )]
  #[Token(
    name: 'enqueue_allowed',
    description: 'Whether the record can be enqueued.',
    classes: [
      SalesforcePullEnqueueEvent::class,
    ],
  )]
  #[Token(
    name: 'operation',
    description: 'The push operation that failed (create, update, or delete).',
    classes: [
      SalesforcePushOpEvent::class,
    ],
  )]
  public function getData(string $key): mixed {
    $event = $this->event;

    // Handle ERROR event tokens
    if ($event instanceof SalesforceErrorEvent) {
      switch ($key) {
        case 'exception':
          return $event->getException();

        case 'error_message':
          return $event->getMessage();

        case 'error_context':
          return $event->getContext();

        case 'log_level':
          return $event->getLevel();

        case 'entity_id':
          $context = $event->getContext();
          return is_array($context) ? ($context['%id'] ?? NULL) : NULL;

        case 'item':
          $context = $event->getContext();
          return is_array($context) ? ($context['%item'] ?? NULL) : NULL;

        case 'mapping_info':
          $context = $event->getContext();
          return is_array($context) ? ($context['%mapping'] ?? NULL) : NULL;

        case 'operation_type':
          $context = $event->getContext();
          return is_array($context) ? ($context['%type'] ?? NULL) : NULL;

        case 'failure_details':
          $context = $event->getContext();
          return is_array($context) ? ($context['%fail'] ?? NULL) : NULL;
      }
    }

    // Provide top-level tokens directly for other events.
    if ($key === 'entity' && method_exists($event, 'getEntity')) {
      return $event->getEntity();
    }

    if ($key === 'mapping' && method_exists($event, 'getMapping')) {
      return $event->getMapping();
    }

    if ($key === 'mapped_object' && method_exists($event, 'getMappedObject')) {
      return $event->getMappedObject();
    }

    // Provide direct aliases for simple properties.
    if ($key === 'mapping_id') {
      $mapping = method_exists($event, 'getMapping') ? $event->getMapping() : NULL;
      return $mapping ? $mapping->id() : NULL;
    }

    if ($key === 'mapping_label') {
      $mapping = method_exists($event, 'getMapping') ? $event->getMapping() : NULL;
      return $mapping ? $mapping->label() : NULL;
    }

    if ($key === 'mapping_salesforce_object_type') {
      $mapping = method_exists($event, 'getMapping') ? $event->getMapping() : NULL;
      return ($mapping && method_exists($mapping, 'getSalesforceObjectType')) ? $mapping->getSalesforceObjectType() : NULL;
    }

    if ($key === 'mapped_object_id') {
      $mappedObject = method_exists($event, 'getMappedObject') ? $event->getMappedObject() : NULL;
      return ($mappedObject && method_exists($mappedObject, 'id')) ? $mappedObject->id() : NULL;
    }

    if ($key === 'mapped_object_salesforce_id') {
      $mappedObject = method_exists($event, 'getMappedObject') ? $event->getMappedObject() : NULL;
      return ($mappedObject && method_exists($mappedObject, 'sfid')) ? $mappedObject->sfid() : NULL;
    }

    if ($key === 'push_allowed' && $event instanceof SalesforcePushAllowedEvent) {
      return $event->isPushAllowed();
    }

    if ($key === 'pull_allowed' && $event instanceof SalesforcePullEvent) {
      return $event->isPullAllowed();
    }

    if ($key === 'params' && $event instanceof SalesforcePushParamsEvent) {
      $pushParams = $event->getParams();
      $params = [];
      if ($pushParams instanceof PushParams) {
        $params = $pushParams->getParams();
      }
      elseif (is_array($pushParams)) {
        $params = $pushParams;
      }
      return $params;
    }

    if ($key === 'query' && $event instanceof SalesforceQueryEvent) {
      $query = $event->getQuery();
      if ($query) {
        return [
          'object_type' => $query->objectType ?? NULL,
          'fields' => $query->fields ?? [],
          'conditions' => $query->conditions ?? [],
          'order' => $query->order ?? [],
          'limit' => $query->limit ?? NULL,
          'offset' => $query->offset ?? NULL,
          'query_string' => (string) $query,
        ];
      }
    }

    if ($key === 'query_string' && $event instanceof SalesforceQueryEvent) {
      $query = $event->getQuery();
      return $query ? (string) $query : NULL;
    }

    if ($key === 'record' && $event instanceof SalesforcePullEnqueueEvent) {
      $record = $event->getRecord();
      if ($record && method_exists($record, 'fields')) {
        return $record->fields();
      }
    }

    if ($key === 'records' && $event instanceof SalesforcePullEnqueueEvent) {
      $queryResult = $event->getRecords();
      $recordsData = [];
      if ($queryResult && method_exists($queryResult, 'records')) {
        $records = $queryResult->records();
        if (is_array($records)) {
          foreach ($records as $idx => $rec) {
            if ($rec && method_exists($rec, 'fields')) {
              $recordsData[$idx] = $rec->fields();
            }
          }
        }
      }
      return $recordsData;
    }

    if ($key === 'force_pull' && $event instanceof SalesforcePullEnqueueEvent) {
      return $event->getForcePull();
    }

    if ($key === 'enqueue_allowed' && $event instanceof SalesforcePullEnqueueEvent) {
      return $event->isEnqueueAllowed();
    }

    if ($key === 'field_plugin' && $event instanceof SalesforcePullEntityValueEvent) {
      $fieldPlugin = $event->getFieldPlugin();
      if (is_object($fieldPlugin)) {
        return [
          'plugin_id' => method_exists($fieldPlugin, 'getPluginId') ? $fieldPlugin->getPluginId() : NULL,
          'configuration' => method_exists($fieldPlugin, 'getConfiguration') ? $fieldPlugin->getConfiguration() : [],
        ];
      }
    }

    if ($key === 'entity_value' && $event instanceof SalesforcePullEntityValueEvent) {
      $entityValue = $event->getEntityValue();
      if (is_object($entityValue) && method_exists($entityValue, '__toString')) {
        return (string) $entityValue;
      }
      elseif (is_object($entityValue) && method_exists($entityValue, 'getValue')) {
        return $entityValue->getValue();
      }
      return $entityValue;
    }

    if ($key === 'operation' && $event instanceof SalesforcePushOpEvent) {
      return $event->getOp();
    }

    return parent::getData($key);
  }

}
